# querygrid.py
# Copyright 2015 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Grids for listing details of selection rules on chess database.
"""

import tkinter.messagebox

from solentware_grid.datagrid import DataGrid

from solentware_misc.gui.exceptionhandler import ExceptionHandler

from ..core.chessrecord import ChessDBrecordQuery
from .querydisplay import QueryDisplay, QueryDisplayEdit
from .queryrow import ChessDBrowQuery
from ..core import exporters
from .eventspec import EventSpec, DummyEvent
from .display import Display


class QueryListGrid(ExceptionHandler, DataGrid, Display):

    """A DataGrid for lists of game selection rules.

    Subclasses provide navigation and extra methods appropriate to list use.

    """

    def __init__(self, parent, ui):
        """Extend with link to user interface object.

        parent - see superclass
        ui - container for user interface widgets and methods.

        """
        super(QueryListGrid, self).__init__(parent=parent)
        self.gcanvas.configure(takefocus=tkinter.FALSE)
        self.data.configure(takefocus=tkinter.FALSE)
        self.frame.configure(takefocus=tkinter.FALSE)
        self.hsbar.configure(takefocus=tkinter.FALSE)
        self.vsbar.configure(takefocus=tkinter.FALSE)
        self.ui = ui
        self.set_event_bindings_frame(
            (
                (EventSpec.tab_traverse_forward, self.traverse_forward),
                (EventSpec.tab_traverse_backward, self.traverse_backward),
                (EventSpec.tab_traverse_round, self.traverse_round),
                # Remove entries when binding implemented in solentware_grid.
                (
                    EventSpec.score_enable_F10_popupmenu_at_top_left,
                    self.show_grid_or_row_popup_menu_at_top_left_by_keypress,
                ),
                (
                    EventSpec.score_enable_F10_popupmenu_at_pointer,
                    self.show_grid_or_row_popup_menu_at_pointer_by_keypress,
                ),
            )
        )

    def display_selected_item(self, key):
        """Create QueryDisplay for game selection rule."""
        selected = self.get_visible_record(key)
        if selected is None:
            return None
        # Should the Frame containing board and position be created here and
        # passed to QueryDisplay. (Needs 'import Tkinter' above.)
        # Rather than passing the container where the Frame created by
        # QueryDisplay is to be put.
        selection = self.make_display_widget(selected)
        self.ui.add_selection_rule_to_display(selection)
        self.ui.selection_items.increment_object_count(key)
        self.ui.selection_items.set_itemmap(selection, key)
        self.set_properties(key)
        return selection

    def make_display_widget(self, sourceobject):
        """Return a QueryDisplay for sourceobject."""
        selection = QueryDisplay(
            master=self.ui.view_selection_rules_pw,
            ui=self.ui,
            items_manager=self.ui.selection_items,
            itemgrid=self.ui.base_games,
            sourceobject=sourceobject,
        )
        selection.query_statement.set_database(
            self.ui.base_games.datasource.dbhome
        )
        selection.query_statement.dbset = self.ui.base_games.datasource.dbset
        selection.query_statement.process_query_statement(
            sourceobject.get_srvalue()
        )
        return selection

    def edit_selected_item(self, key):
        """Create a QueryDisplayEdit for game selection rule."""
        selected = self.get_visible_record(key)
        if selected is None:
            return None
        # Should the Frame containing board and position be created here and
        # passed to QueryDisplayEdit. (Which needs 'import Tkinter' above.)
        # Rather than passing the container where the Frame created by
        # QueryDisplayEdit is to be put.
        selection = self.make_edit_widget(selected)
        self.ui.add_selection_rule_to_display(selection)
        self.ui.selection_items.increment_object_count(key)
        self.ui.selection_items.set_itemmap(selection, key)
        self.set_properties(key)
        return selection

    def make_edit_widget(self, sourceobject):
        """Return a QueryDisplayEdit for sourceobject."""
        selection = QueryDisplayEdit(
            master=self.ui.view_selection_rules_pw,
            ui=self.ui,
            items_manager=self.ui.selection_items,
            itemgrid=self.ui.base_games,
            sourceobject=sourceobject,
        )
        selection.query_statement.set_database(
            self.ui.base_games.datasource.dbhome
        )
        selection.query_statement.dbset = self.ui.base_games.datasource.dbset
        selection.query_statement.process_query_statement(
            sourceobject.get_srvalue()
        )
        return selection

    def set_properties(self, key, dodefaultaction=True):
        """Return True if properties for selection rule key set or False."""
        if super(QueryListGrid, self).set_properties(
            key, dodefaultaction=False
        ):
            return True
        if self.ui.selection_items.object_display_count(key):
            self.objects[key].set_background_on_display(
                self.get_row_widgets(key)
            )
            self.set_row_under_pointer_background(key)
            return True
        if dodefaultaction:
            self.objects[key].set_background_normal(self.get_row_widgets(key))
            self.set_row_under_pointer_background(key)
            return True
        return False

    def set_row(self, key, dodefaultaction=True, **kargs):
        """Return row widget for selection rule key or None."""
        row = super(QueryListGrid, self).set_row(
            key, dodefaultaction=False, **kargs
        )
        if row is not None:
            return row
        if key not in self.keys:
            return None
        if self.ui.selection_items.object_display_count(key):
            return self.objects[key].grid_row_on_display(**kargs)
        if dodefaultaction:
            return self.objects[key].grid_row_normal(**kargs)
        else:
            return None

    def select_down(self):
        """Extend to show selection summary in status bar."""
        super(QueryListGrid, self).select_down()
        self.set_selection_text()

    def select_up(self):
        """Extend to show selection summary in status bar."""
        super(QueryListGrid, self).select_up()
        self.set_selection_text()

    def cancel_selection(self):
        """Extend to clear selection summary from status bar."""
        if self.selection:
            self.ui.statusbar.set_status_text("")
        super(QueryListGrid, self).cancel_selection()

    def launch_delete_record(self, key, modal=True):
        """Create delete dialogue."""
        oldobject = ChessDBrecordQuery()
        oldobject.set_database(self.ui.base_games.datasource.dbhome)
        oldobject.value.dbset = self.ui.base_games.datasource.dbset
        oldobject.load_record(
            (self.objects[key].key.pack(), self.objects[key].srvalue)
        )
        self.create_delete_dialog(
            self.objects[key], oldobject, modal, title="Delete Selection Rule"
        )

    def launch_edit_record(self, key, modal=True):
        """Create edit dialogue."""
        self.create_edit_dialog(
            self.objects[key],
            ChessDBrecordQuery(),
            ChessDBrecordQuery(),
            False,
            modal,
            title="Edit Selection Rule",
        )

    def launch_edit_show_record(self, key, modal=True):
        """Create edit dialogue including reference copy of original."""
        self.create_edit_dialog(
            self.objects[key],
            ChessDBrecordQuery(),
            ChessDBrecordQuery(),
            True,
            modal,
            title="Edit Selection Rule",
        )

    def launch_insert_new_record(self, modal=True):
        """Create insert dialogue."""
        instance = self.datasource.new_row()

        # Later process_query_statement() causes display of empty title and
        # query lines.
        instance.srvalue = repr("")

        self.create_edit_dialog(
            instance,
            ChessDBrecordQuery(),
            None,
            False,
            modal,
            title="New Selection Rule",
        )

    def launch_show_record(self, key, modal=True):
        """Create show dialogue."""
        oldobject = ChessDBrecordQuery()
        oldobject.set_database(self.ui.base_games.datasource.dbhome)
        oldobject.value.dbset = self.ui.base_games.datasource.dbset
        oldobject.load_record(
            (self.objects[key].key.pack(), self.objects[key].srvalue)
        )
        self.create_show_dialog(
            self.objects[key], oldobject, modal, title="Show Selection Rule"
        )

    def create_edit_dialog(
        self, instance, newobject, oldobject, showinitial, modal, title=""
    ):
        """Extend to do chess initialization."""
        for x in (newobject, oldobject):
            if x:
                x.set_database(self.ui.base_games.datasource.dbhome)
                x.value.dbset = self.ui.base_games.datasource.dbset
                x.load_record((instance.key.pack(), instance.srvalue))
        super(QueryListGrid, self).create_edit_dialog(
            instance, newobject, oldobject, showinitial, modal, title=title
        )

    def fill_view(
        self,
        currentkey=None,
        down=True,
        topstart=True,
        exclude=True,
    ):
        """Delegate to superclass if database is open otherwise do nothing."""

        # Intend to put this in superclass but must treat the DataClient objects
        # being scrolled as a database to do this properly.  Do this when these
        # objects have been given a database interface used when the database
        # is not open.  (One problem is how to deal with indexes.)

        # Used to deal with temporary closure of database to do Imports of games
        # from PGN files; which can take many hours.

        if self.get_database() is not None:
            super(QueryListGrid, self).fill_view(
                currentkey=currentkey,
                down=down,
                topstart=topstart,
                exclude=exclude,
            )

    def load_new_index(self):
        """Delegate to superclass if database is open otherwise do nothing."""

        # Intend to put this in superclass but must treat the DataClient objects
        # being scrolled as a database to do this properly.  Do this when these
        # objects have been given a database interface used when the database
        # is not open.  (One problem is how to deal with indexes.)

        # Used to deal with temporary closure of database to do Imports of games
        # from PGN files; which can take many hours.

        if self.get_database() is not None:
            super(QueryListGrid, self).load_new_index()

    def load_new_partial_key(self, key):
        """Delegate to superclass if database is open otherwise do nothing."""

        # Intend to put this in superclass but must treat the DataClient objects
        # being scrolled as a database to do this properly.  Do this when these
        # objects have been given a database interface used when the database
        # is not open.  (One problem is how to deal with indexes.)

        # Used to deal with temporary closure of database to do Imports of games
        # from PGN files; which can take many hours.

        if self.get_database() is not None:
            super(QueryListGrid, self).load_new_partial_key(key)

    def on_configure_canvas(self, event=None):
        """Delegate to superclass if database is open otherwise do nothing."""

        # Intend to put this in superclass but must treat the DataClient objects
        # being scrolled as a database to do this properly.  Do this when these
        # objects have been given a database interface used when the database
        # is not open.  (One problem is how to deal with indexes.)

        # Used to deal with temporary closure of database to do Imports of games
        # from PGN files; which can take many hours.

        if self.get_database() is not None:
            super(QueryListGrid, self).on_configure_canvas(event=event)

    def on_data_change(self, instance):
        """Delegate to superclass if database is open otherwise do nothing."""

        # Intend to put this in superclass but must treat the DataClient objects
        # being scrolled as a database to do this properly.  Do this when these
        # objects have been given a database interface used when the database
        # is not open.  (One problem is how to deal with indexes.)

        # Used to deal with temporary closure of database to do Imports of games
        # from PGN files; which can take many hours.

        if self.get_database() is not None:
            super(QueryListGrid, self).on_data_change(instance)

    def set_popup_bindings(self, popup, bindings=()):
        for accelerator, function in bindings:
            popup.add_command(
                label=accelerator[1],
                command=self.try_command(function, popup),
                accelerator=accelerator[2],
            )

    def add_cascade_menu_to_popup(self, index, popup, bindings=None):
        """Add cascade_menu, and bindings, to popup if not already present.

        The index is used as the label on the popup menu when visible.

        The bindings are not applied if cascade_menu is alreay in popup menu.

        """
        # Cannot see a way of asking 'Does entry exist?' other than:
        try:
            popup.index(index)
        except:
            cascade_menu = tkinter.Menu(master=popup, tearoff=False)
            popup.add_cascade(label=index, menu=cascade_menu)
            if bindings is None:
                return
            self.set_popup_bindings(cascade_menu, bindings)

    def set_event_bindings_frame(self, bindings=(), switch=True):
        """Set bindings if switch is True or unset the bindings."""
        ste = self.try_event
        for sequence, function in bindings:
            self.frame.bind(
                sequence[0], ste(function) if switch and function else ""
            )

    def traverse_backward(self, event=None):
        """Give focus to previous widget type in traversal order."""
        self.ui.give_focus_backward(self)
        return "break"

    def traverse_forward(self, event=None):
        """Give focus to next widget type in traversal order."""
        self.ui.give_focus_forward(self)
        return "break"

    def traverse_round(self, event=None):
        """Give focus to next widget within active item in traversal order."""
        return "break"

    def set_focus(self):
        """Give focus to this widget."""
        self.frame.focus_set()
        if self.ui.single_view:
            self.ui.show_just_panedwindow_with_focus(self.frame)

    def is_payload_available(self):
        """Return True if grid is connected to a database."""
        ds = self.get_data_source()
        if ds is None:
            return False
        if ds.get_database() is None:

            # Avoid exception scrolling visible grid not connected to database.
            # Make still just be hack to cope with user interface activity
            # while importing data.
            self.clear_grid_keys()

            return False
        return True

    def export_selection_rules(self, event=None):
        """Export selected selection rule definitions."""
        exporters.export_selected_selection_rules(
            self, self.ui.get_export_filename("Selection Rules", pgn=False)
        )

    def focus_set_frame(self, event=None):
        """Adjust widget which is losing focus then delegate to superclass."""
        self.ui.set_bindings_on_item_losing_focus_by_pointer_click()
        super().focus_set_frame(event=event)

    def bind_for_widget_without_focus(self):
        """Return True if this item has the focus about to be lost."""
        if self.get_frame().focus_displayof() != self.get_frame():
            return False

        # Nothing to do on losing focus.
        return True

    def get_top_widget(self):
        """Return topmost widget for game display.

        The topmost widget is put in a container widget in some way.

        """
        # Superclass DataGrid.get_frame() method returns the relevant widget.
        # Name, get_top_widget, is compatible with Game and Partial names.
        return self.get_frame()

    def get_visible_selected_key(self):
        """Return selected key if it is visible and display dialogue if not.

        Getting the key is delegated to superclass.

        """
        key = super().get_visible_selected_key()
        if key is None:
            tkinter.messagebox.showinfo(
                parent=self.parent,
                title="Display Item",
                message="No record selected or selected record is not visible",
            )
        return key


class QueryGrid(QueryListGrid):

    """Customized QueryListGrid for list of game selection rules."""

    def __init__(self, ui):
        """Extend with definition and bindings for selection rules on grid.

        ui - container for user interface widgets and methods.

        """
        super(QueryGrid, self).__init__(ui.selection_rules_pw, ui)
        self.make_header(ChessDBrowQuery.header_specification)
        self.__bind_on()
        self.set_popup_bindings(
            self.menupopup,
            (
                (
                    EventSpec.display_record_from_grid,
                    self.display_selection_rule_from_popup,
                ),
                (
                    EventSpec.edit_record_from_grid,
                    self.edit_selection_rule_from_popup,
                ),
            ),
        )
        bindings = (
            (
                EventSpec.navigate_to_position_grid,
                self.set_focus_position_grid,
            ),
            (
                EventSpec.navigate_to_active_game,
                self.set_focus_gamepanel_item_command,
            ),
            (EventSpec.navigate_to_game_grid, self.set_focus_game_grid),
            (
                EventSpec.navigate_to_repertoire_grid,
                self.set_focus_repertoire_grid,
            ),
            (
                EventSpec.navigate_to_active_repertoire,
                self.set_focus_repertoirepanel_item_command,
            ),
            (
                EventSpec.navigate_to_repertoire_game_grid,
                self.set_focus_repertoire_game_grid,
            ),
            (
                EventSpec.navigate_to_active_partial,
                self.set_focus_partialpanel_item_command,
            ),
            (
                EventSpec.navigate_to_partial_game_grid,
                self.set_focus_partial_game_grid,
            ),
            (EventSpec.navigate_to_partial_grid, self.set_focus_partial_grid),
            (
                EventSpec.navigate_to_active_selection_rule,
                self.set_focus_selectionpanel_item_command,
            ),
            (EventSpec.tab_traverse_backward, self.traverse_backward),
            (EventSpec.tab_traverse_forward, self.traverse_forward),
        )
        self.add_cascade_menu_to_popup("Navigation", self.menupopup, bindings)
        self.add_cascade_menu_to_popup(
            "Navigation", self.menupopupnorow, bindings
        )

    def bind_off(self):
        """Disable all bindings."""
        super(QueryGrid, self).bind_off()
        self.set_event_bindings_frame(
            (
                (EventSpec.navigate_to_active_partial, ""),
                (EventSpec.navigate_to_partial_game_grid, ""),
                (EventSpec.navigate_to_repertoire_grid, ""),
                (EventSpec.navigate_to_active_repertoire, ""),
                (EventSpec.navigate_to_repertoire_game_grid, ""),
                (EventSpec.navigate_to_position_grid, ""),
                (
                    EventSpec.navigate_to_active_game,
                    self.set_focus_gamepanel_item,
                ),
                (EventSpec.navigate_to_game_grid, ""),
                (EventSpec.navigate_to_partial_grid, ""),
                (EventSpec.navigate_to_active_selection_rule, ""),
                (EventSpec.display_record_from_grid, ""),
                (EventSpec.edit_record_from_grid, ""),
            )
        )

    def bind_on(self):
        """Enable all bindings."""
        super(QueryGrid, self).bind_on()
        self.__bind_on()

    def __bind_on(self):
        """Enable all bindings."""
        self.set_event_bindings_frame(
            (
                (
                    EventSpec.navigate_to_active_partial,
                    self.set_focus_partialpanel_item,
                ),
                (
                    EventSpec.navigate_to_partial_game_grid,
                    self.set_focus_partial_game_grid,
                ),
                (
                    EventSpec.navigate_to_repertoire_grid,
                    self.set_focus_repertoire_grid,
                ),
                (
                    EventSpec.navigate_to_active_repertoire,
                    self.set_focus_repertoirepanel_item,
                ),
                (
                    EventSpec.navigate_to_repertoire_game_grid,
                    self.set_focus_repertoire_game_grid,
                ),
                (
                    EventSpec.navigate_to_position_grid,
                    self.set_focus_position_grid,
                ),
                (
                    EventSpec.navigate_to_active_game,
                    self.set_focus_gamepanel_item,
                ),
                (EventSpec.navigate_to_game_grid, self.set_focus_game_grid),
                (
                    EventSpec.navigate_to_partial_grid,
                    self.set_focus_partial_grid,
                ),
                (
                    EventSpec.navigate_to_active_selection_rule,
                    self.set_focus_selectionpanel_item,
                ),
                (
                    EventSpec.display_record_from_grid,
                    self.display_selection_rule,
                ),
                (EventSpec.edit_record_from_grid, self.edit_selection_rule),
            )
        )

    def display_selection_rule(self, event=None):
        """Display selection rule and cancel selection.

        Call _display_selection_rule after idle tasks to allow message display.

        """
        if not self.get_visible_selected_key():
            return
        self._set_find_selection_rule_name_games(self.selection[0])
        self.frame.after_idle(
            self.try_command(self._display_selection_rule, self.frame)
        )

    def display_selection_rule_from_popup(self, event=None):
        """Display selection rule selected by pointer.

        Call _display_selection_rule after idle tasks to allow message display.

        """
        self._set_find_selection_rule_name_games(self.pointer_popup_selection)
        self.frame.after_idle(
            self.try_command(
                self._display_selection_rule_from_popup, self.frame
            )
        )

    def _display_selection_rule(self):
        """Display selection rule and cancel selection.

        Call from display_selection_rule only.

        """
        self.display_selected_item(self.get_visible_selected_key())
        self.cancel_selection()

    def _display_selection_rule_from_popup(self):
        """Display selection rule selected by pointer.

        Call from display_selection_rule_from_popup only.

        """
        self.display_selected_item(self.pointer_popup_selection)

    def edit_selection_rule(self, event=None):
        """Display selection rule allow editing and cancel selection.

        Call _edit_selection_rule after idle tasks to allow message display.

        """
        if not self.get_visible_selected_key():
            return
        self._set_find_selection_rule_name_games(self.selection[0])
        self.frame.after_idle(
            self.try_command(self._edit_selection_rule, self.frame)
        )

    def edit_selection_rule_from_popup(self, event=None):
        """Display selection rule with editing allowed selected by pointer.

        Call _edit_selection_rule after idle tasks to allow message display.

        """
        self._set_find_selection_rule_name_games(self.pointer_popup_selection)
        self.frame.after_idle(
            self.try_command(self._edit_selection_rule_from_popup, self.frame)
        )

    def _edit_selection_rule(self):
        """Display selection rule allow editing and cancel selection.

        Call from edit_selection_rule only.

        """
        self.edit_selected_item(self.get_visible_selected_key())
        self.cancel_selection()

    def _edit_selection_rule_from_popup(self):
        """Display selection rule with editing allowed selected by pointer.

        Call from edit_selection_rule_from_popup only.

        """
        self.edit_selected_item(self.pointer_popup_selection)

    def _set_find_selection_rule_name_games(self, key):
        """Set status text to active selection rule name."""
        if self.ui.selection_items.count_items_in_stack():
            # do search at this time only if no selection rules displayed
            return
        self.ui.statusbar.set_status_text(
            "".join(
                (
                    "Please wait while finding games for selection rule ",
                    self.objects[key].value.get_name_text(),
                )
            )
        )

    def on_partial_change(self, instance):
        # may turn out to be just to catch datasource is None
        if self.get_data_source() is None:
            return
        super(QueryGrid, self).on_data_change(instance)

    def set_selection_text(self):
        """Set status bar to display selection rule name."""
        if self.selection:
            p = self.objects[self.selection[0]].value
            self.ui.statusbar.set_status_text(
                "".join(
                    (
                        p.get_name_text(),
                        "   (",
                        p.get_query_statement_text(),
                        ")",
                    )
                )
            )
        else:
            self.ui.statusbar.set_status_text("")

    def is_visible(self):
        """Return True if list of selection rules is displayed."""
        return str(self.get_frame()) in self.ui.selection_rules_pw.panes()

    def make_display_widget(self, sourceobject):
        """Return a QueryDisplay for sourceobject."""
        selection = super().make_display_widget(sourceobject)
        selection.set_and_tag_item_text()
        return selection

    def make_edit_widget(self, sourceobject):
        """Return a QueryDisplayEdit for sourceobject."""
        selection = super().make_edit_widget(sourceobject)
        selection.set_and_tag_item_text(reset_undo=True)
        return selection

    def focus_set_frame(self, event=None):
        """Delegate to superclass then set toolbar widget states."""
        super().focus_set_frame(event=event)
        self.ui.set_toolbarframe_normal(
            self.ui.move_to_selection, self.ui.filter_selection
        )

    def set_selection(self, key):
        """Hack to fix edge case when inserting records using apsw or sqlite3.

        Workaround a KeyError exception when a record is inserted while a grid
        keyed by a secondary index with only one key value in the index is on
        display.

        """
        try:
            super().set_selection(key)
        except KeyError:
            tkinter.messagebox.showinfo(
                parent=self.parent,
                title="Insert Selection Rule Workaround",
                message="".join(
                    (
                        "All records have same name on this display.\n\nThe new ",
                        "record has been inserted but you need to Hide, and then ",
                        "Show, the display to see the record in the list.",
                    )
                ),
            )



# querytoplevel.py
# Copyright 2021 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Toplevel widgets to display and edit game selection rules.

These two classes display game selection rules in their own Toplevel widget:
they are used in the querydbdelete, querydbedit, and querydbshow, modules.

"""

from .query import Query
from .queryedit import QueryEdit
from .topleveltext import ToplevelText


class QueryToplevel(ToplevelText, Query):

    """Customize Query to be the single instance in a Toplevel widget."""


class QueryToplevelEdit(ToplevelText, QueryEdit):

    """Customize QueryEdit to be the single instance in a Toplevel widget."""

    # A method like GameToplevelEdit.create_primary_activity_popup is not
    # needed because the standard edit operations of Text widget are
    # sufficient.



# cqlds.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Interface to chess database partial position index using ChessQL.

The ChessQueryLanguageDS class represents a subset of games which match a Chess
Query Language query.

The ChessQueryLanguageDS class in this module supports the DPT interface to a
database.

See the ..basecore.cqlds module for the ChessQueryLanguageDS class for apsw,
db, and sqlite3.

"""

from solentware_grid.dpt.datasourcecursor import DataSourceCursor

from ..basecore.cqlgames import ChessQLGames


class ChessQueryLanguageDS(ChessQLGames, DataSourceCursor):

    """Combine a DPT DataSourceCursor with ChessQLGames."""



# cqldbdelete.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Customise delete toplevel to delete Chess Query Language (ChessQL) statement
record.

ChessQL statements obey the syntax published for CQL version 6.0.1 (by Gady
Costeff).

"""
from solentware_grid.gui.datadelete import DataDelete
from solentware_misc.gui.exceptionhandler import ExceptionHandler

from .cqltoplevel import CQLToplevel
from .topleveltext import DeleteText


class CQLDbDelete(ExceptionHandler, DeleteText, DataDelete):
    """Delete ChessQL statement from database.

    parent is used as the master argument in a CQLToplevel call.

    ui is used as the ui argument in a CQLToplevel call.

    parent, oldobject, and the CQLToplevel instance created, are used as
    arguments in the super.__init__ call.

    Attribute text_name provides the name used in widget titles and message
    text.

    Methods get_title_for_object and set_item, and properties ui_base_table;
    ui_items_in_toplevels; and ui, allow similar methods in various classes
    to be expressed identically and defined once.

    """

    text_name = "ChessQL Statement"

    def __init__(self, parent, oldobject, ui=None):
        """Extend and create toplevel widget to delete ChessQL statement."""
        # Toplevel title set '' in __init__ and to proper value in initialize.
        super().__init__(
            oldobject, parent, CQLToplevel(master=parent, ui=ui), ""
        )
        self.initialize()

    def get_title_for_object(self, object_=None):
        """Return title for Toplevel containing a ChessQL statement object_.

        Default value of object_ is object attribute from DataDelete class.

        """
        if object_ is None:
            object_ = self.object
        return "  ".join(
            (
                self.text_name.join(("Delete ", ":")),
                object_.value.get_name_text(),
            )
        )

    @property
    def ui_base_table(self):
        return self.ui.base_partials

    @property
    def ui_items_in_toplevels(self):
        return self.ui.partials_in_toplevels

    @property
    def ui(self):
        return self.oldview.ui

    def set_item(self, view, object_):
        view.cql_statement.process_statement(object_.get_srvalue())
        view.set_and_tag_item_text()

    def dialog_ok(self):
        """Return delete action response (True for deleted).

        Delegate to superclass if the active item on the main display is not a
        'Position | Partial' widget and is not set to edit the same record.

        """
        if self.ui.partial_items.active_item:
            if self.ui.partial_items.active_item.sourceobject is None:
                tkinter.messagebox.showinfo(
                    parent=self.parent,
                    title=self.get_title_for_object(),
                    message="".join(
                        (
                            "Cannot use this delete dialogue while the active ",
                            "item in cql queries is one opened by menu action ",
                            "'Position | Partial'.",
                        )
                    ),
                )
                return False
            if (
                self.ui.partial_items.active_item.sourceobject.key
                == self.object.key
            ):
                tkinter.messagebox.showinfo(
                    parent=self.parent,
                    title=self.get_title_for_object(),
                    message="".join(
                        (
                            "Cannot use this delete dialogue while the active ",
                            "item in cql queries is one that displays the record ",
                            "being deleted.",
                        )
                    ),
                )
                return False
        return super().dialog_ok()

    def delete(self, commit=True):
        """Delegate to superclass to delete record then delete game list."""
        if commit:
            self.datasource.dbhome.start_transaction()
        super().delete(commit=False)
        cqls = self.ui.partialpositionds(
            self.ui.base_games.datasource.dbhome,
            self.ui.base_games.datasource.dbset,
            self.ui.base_games.datasource.dbset,
            newrow=None,
        )
        assert self.object.newrecord is None
        cqls.forget_cql_statement_games(self.object, commit=False)
        if commit:
            self.datasource.dbhome.commit()



# repertoireedit.py
# Copyright 2008, 2021 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Widget to edit an opening repertoire.

The RepertoireEdit class displays PGN text representing an opening repertoire
and allows editing.  It is a subclass of GameEdit.

This class does not allow deletion of repertoires from a database.

An instance of these classes fits into the user interface in two ways: as an
item in a panedwindow of the main widget, or as the only item in a new toplevel
widget.

"""

from ..core.constants import REPERTOIRE_TAG_ORDER
from ..core.pgn import GameDisplayMoves, GameRepertoireDisplayMoves
from .eventspec import EventSpec
from .gameedit import GameEdit


class RepertoireEdit(GameEdit):

    """Display a repertoire with editing allowed.

    gameclass is passed to the superclass as the gameclass argument.  It
    defaults to GameDisplayMoves.

    Attribute tags_displayed_last is the PGN tags, in order, to be displayed
    immediately before the movetext.  It exists so Game*, Repertoire*, and
    AnalysisScore*, instances can use identical code to display PGN tags.  It
    is the PGN repertoire tags defined in ChessTab.

    Attribute pgn_export_type is a tuple with the name of the type of data and
    the class used to generate export PGN.  It exists so Game*, Repertoire*,
    and AnalysisScore*, instances can use identical code to display PGN tags.
    It is ('Repertoire', GameRepertoireDisplayMoves).

    """

    tags_displayed_last = REPERTOIRE_TAG_ORDER
    pgn_export_type = "Repertoire", GameRepertoireDisplayMoves

    # gameclass=GameRepertoireDisplayMoves surely?
    # Then maybe do not need pgn_export_type for 'export_..' methods in Score.
    # Otherwise there is no point to this __init__ method.
    def __init__(self, gameclass=GameDisplayMoves, **ka):
        """Extend with bindings to edit repertoire score."""
        super(RepertoireEdit, self).__init__(gameclass=gameclass, **ka)

    def insert_empty_pgn_seven_tag_roster(self):
        """Insert ' [ <fieldname> "<null>" ... ] ' seven tag roster sequence."""
        self.set_insertion_point_before_next_pgn_tag()
        for t in REPERTOIRE_TAG_ORDER:
            self.add_pgntag_to_map(t, "")

    # There is no point to a repertoire without RAVs so the options suppressing
    # RAVs are absent.
    def get_all_export_events(self):
        return (
            (
                EventSpec.pgn_export_format_no_comments,
                self.export_pgn_no_comments,
            ),
            (EventSpec.pgn_export_format, self.export_pgn),
            (EventSpec.pgn_import_format, self.export_pgn_import_format),
            (EventSpec.text_internal_format, self.export_text),
        )



# datarow.py
# Copyright 2020 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Add solentware_misc exception handling to solentware_grid DataRow class for
ChessTab.
"""

from solentware_grid.gui.datarow import DataRow

from solentware_misc.gui.exceptionhandler import ExceptionHandler


class DataRow(ExceptionHandler, DataRow):

    """Override solentware_grid.gui.datarow.DataRow methods, from the subclass
    CallbackException, with those in ExceptionHandler.
    """



# cqlds.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Interface to chess database partial position index using ChessQL.

The ChessQueryLanguageDS class represents a subset of games which match a Chess
Query Language query.

The ChessQueryLanguageDS class in this module supports the apsw, db, and
sqlite3, interfaces to a database.

See the ..dpt.cqlds module for the ChessQueryLanguageDS class for DPT.

"""

from solentware_grid.core.datasourcecursor import DataSourceCursor

from .cqlgames import ChessQLGames


class ChessQueryLanguageDS(ChessQLGames, DataSourceCursor):

    """Combine a standard DataSourceCursor with ChessQLGames."""



# enginedbshow.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Customise show toplevel to show chess engine definition record.
"""

from solentware_grid.gui.datashow import DataShow

from solentware_misc.gui.exceptionhandler import ExceptionHandler

from .enginetoplevel import EngineToplevel
from .topleveltext import ShowText


class EngineDbShow(ExceptionHandler, ShowText, DataShow):
    """Show a chess engine definition from database.

    parent is used as the master argument in an EngineToplevel call.

    ui is used as the ui argument in an EngineToplevel call.

    parent, oldobject, and the EngineToplevel instance created, are used as
    arguments in the super.__init__ call.

    Attribute text_name provides the name used in widget titles and message
    text.

    Methods get_title_for_object and set_item, and properties ui_base_table;
    ui_items_in_toplevels; and ui, allow similar methods in various classes
    to be expressed identically and defined once.

    """

    text_name = "Engine Definition"

    def __init__(self, parent, oldobject, ui=None):
        """Create toplevel widget for showing chess engine definition.

        ui should be a UCI instance.

        """
        # Toplevel title set '' in __init__ and to proper value in initialize.
        super().__init__(
            oldobject, parent, EngineToplevel(master=parent, ui=ui), ""
        )
        self.initialize()

    def get_title_for_object(self, object_=None):
        """Return title for Toplevel containing a chess engine definition
        object_.

        Default value of object_ is object attribute from DataShow class.

        """
        if object_ is None:
            object_ = self.object
        return "  ".join(
            (
                self.text_name.join(("Show ", ":")),
                object_.value.get_name_text(),
            )
        )

    @property
    def ui_base_table(self):
        return self.ui.base_engines

    @property
    def ui_items_in_toplevels(self):
        return self.ui.engines_in_toplevels

    @property
    def ui(self):
        return self.oldview.ui

    def set_item(self, view, object_):
        view.definition.extract_engine_definition(object_.get_srvalue())
        view.set_engine_definition(object_.value)

    def tidy_on_destroy(self):
        # ui_base_table is None when this happens other than directly closing
        # the Toplevel.
        try:
            super().tidy_on_destroy()
        except AttributeError:
            if self.ui_base_table is not None:
                raise



# rayfilter.py
# Copyright 2017 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess Query Language (ChessQL) ray filter evaluator.

Examples of ray filters are 'ray ( Q n k )' and 'ray ( Q[a4,c3] n kh1-8 )'.

RayFilter expands the list of square descriptions into the particular rays,
horizontal, vertical, and diagonal, which need to be evaluated.

"""
from chessql.core import constants
from chessql.core import piecedesignator
from chessql.core.cql import Token

from ..core.constants import MOVE_NUMBER_KEYS
from ..core import filespec

# Longest line is eight squares.  Two end points give a maximum of six internal
# squares.  Shorter lines drop elements from the right.  May add a seventh zero
# element to avoid a len() < 8 test to work around an index error exception.
# ray ( Q n b r ) for lines of length five uses MAP_RAY_TO_LINE[2][:][:3] where
# the [:3] part contains two non-zero elements.  These are the lines with one
# empty internal square.
# Non-zero elements in MAP_RAY_TO_LINE[n][m] are the internal piece designator
# in the ray.
MAP_RAY_TO_LINE = [
    [[0, 0, 0, 0, 0, 0]],  # ray ( a A )
    [
        [1, 0, 0, 0, 0, 0],  # ray ( a A K )
        [0, 1, 0, 0, 0, 0],
        [0, 0, 1, 0, 0, 0],
        [0, 0, 0, 1, 0, 0],
        [0, 0, 0, 0, 1, 0],
        [0, 0, 0, 0, 0, 1],
    ],
    [
        [1, 2, 0, 0, 0, 0],  # ray ( a Q b K )
        [1, 0, 2, 0, 0, 0],
        [0, 1, 2, 0, 0, 0],
        [1, 0, 0, 2, 0, 0],
        [0, 1, 0, 2, 0, 0],
        [0, 0, 1, 2, 0, 0],
        [1, 0, 0, 0, 2, 0],
        [0, 1, 0, 0, 2, 0],
        [0, 0, 1, 0, 2, 0],
        [0, 0, 0, 1, 2, 0],
        [1, 0, 0, 0, 0, 2],
        [0, 1, 0, 0, 0, 2],
        [0, 0, 1, 0, 0, 2],
        [0, 0, 0, 1, 0, 2],
        [0, 0, 0, 0, 1, 2],
    ],
    [
        [1, 2, 3, 0, 0, 0],  # ray ( a Q b N K )
        [1, 2, 0, 3, 0, 0],
        [1, 0, 2, 3, 0, 0],
        [0, 1, 2, 3, 0, 0],
        [1, 2, 0, 0, 3, 0],
        [1, 0, 2, 0, 3, 0],
        [1, 0, 0, 2, 3, 0],
        [0, 1, 2, 0, 3, 0],
        [0, 0, 1, 2, 3, 0],
        [1, 2, 0, 0, 0, 3],
        [1, 0, 2, 0, 0, 3],
        [1, 0, 0, 2, 0, 3],
        [1, 0, 0, 0, 2, 3],
        [0, 1, 2, 0, 0, 3],
        [0, 0, 1, 2, 0, 3],
        [0, 0, 1, 0, 2, 3],
        [0, 0, 0, 1, 2, 3],
    ],
    [
        [1, 2, 3, 4, 0, 0],  # ray ( a Q b N p K )
        [1, 2, 3, 0, 4, 0],
        [1, 2, 0, 3, 4, 0],
        [1, 0, 2, 3, 4, 0],
        [0, 1, 2, 3, 4, 0],
        [1, 2, 3, 0, 0, 4],
        [1, 2, 0, 3, 0, 4],
        [1, 2, 0, 0, 3, 4],
        [1, 0, 2, 3, 0, 4],
        [1, 0, 2, 0, 3, 4],
        [1, 0, 0, 2, 3, 4],
        [0, 1, 2, 3, 0, 4],
        [0, 1, 2, 0, 3, 4],
        [0, 1, 0, 2, 3, 4],
        [0, 0, 1, 2, 3, 4],
    ],
    [
        [1, 2, 3, 4, 5, 0],  # ray ( a Q b N p p K )
        [1, 2, 3, 4, 0, 5],
        [1, 2, 3, 0, 4, 5],
        [1, 2, 0, 3, 4, 5],
        [1, 0, 2, 3, 4, 5],
        [0, 1, 2, 3, 4, 5],
    ],
    [[1, 2, 3, 4, 5, 6]],  # ray ( a Q b N p p N K )
]


class RayFilterError(Exception):
    pass


class RayFilter:
    """ChessQL ray filter evaluator.

    The ray and between filters have a list of square specifiers, usually piece
    designators, which define the rays to be evaluated.

    This class assumes the caller has expanded the piece designator parameters
    to the ray or between filter; and applied any transforms.

    Use this class like:
    C(RayFilter)

    Subclasses must implement the database interface specific methods defined
    in this class which raise RayFilterError('Not implemented')
    exceptions.

    """

    def __init__(self, filter_, move_number, variation_code):
        """ """
        if filter_.tokendef not in {Token.BETWEEN, Token.RAY}:
            raise RayFilterError(
                "".join(
                    (
                        "Filter '",
                        filter_.name,
                        "' does not support rays.",
                    )
                )
            )

        # Is this really needed!
        if len(filter_.children) != 1:
            raise RayFilterError(
                "".join(
                    (
                        "Filter '",
                        filter_.name,
                        "' format not correct.",
                    )
                )
            )
        if filter_.children[0].tokendef is not Token.LEFTPARENTHESIS:
            raise RayFilterError(
                "".join(
                    (
                        "Filter '",
                        filter_.name,
                        "' format not correct.",
                    )
                )
            )

        self.move_number = move_number
        self.variation_code = variation_code
        raycomponents = []
        mvi = move_number_str(move_number) + variation_code
        for c in filter_.children[0].children:
            designator_set = set()
            raycomponents.append(designator_set)
            stack = [c]
            while stack:
                if stack[-1].tokendef is Token.PIECE_DESIGNATOR:
                    designator_set.update(
                        piece_square_to_index(
                            stack[-1].data.designator_set, mvi
                        )
                    )
                    stack.pop()
                    continue
                sp = stack.pop()
                for spc in sp.children:
                    stack.append(spc)
        self.raycomponents = raycomponents
        self.emptycomponents = [set() for i in range(len(raycomponents))]
        self.empty_square_games = set()
        self.piece_square_games = set()
        self.recordset_cache = {}
        self.ray_games = {}

    def prune_end_squares(self, database):
        """Remove ray-end squares with no game references"""
        anypiece = (
            constants.ANY_WHITE_PIECE_NAME + constants.ANY_BLACK_PIECE_NAME
        )
        nopiece = constants.EMPTY_SQUARE_NAME
        fd = filespec.PIECESQUAREMOVE_FIELD_DEF, filespec.SQUAREMOVE_FIELD_DEF
        values_finder = database.values_finder(finder.dbset)
        move = move_number_str(self.move_number)
        nextmove = move_number_str(self.move_number + 1)
        psmwhere, smwhere = [
            database.values_selector(
                " ".join((f, "from", move, "below", nextmove))
            )
            for f in fd
        ]
        for w in psmwhere, smwhere:
            w.lex()
            w.parse()
            w.evaluate(values_finder)
        moveindex = set(psmwhere.node.result + smwhere.node.result)
        for end in 0, -1:
            if nopiece in "".join(self.raycomponents[end]):
                emptyset = self.emptycomponents[end]
                empty = [
                    s[:-1] for s in self.raycomponents[end] if nopiece in s
                ]
                for p in anypiece:
                    for e in empty:
                        if e + p in moveindex:
                            continue
                        emptyset.add(e)
            self.raycomponents[end].intersection_update(moveindex)

    def find_games_for_end_squares(self, finder):
        """Remove ray-end squares with no game references"""
        anywhitepiece = constants.ANY_WHITE_PIECE_NAME
        anyblackpiece = constants.ANY_BLACK_PIECE_NAME
        anypiece = anywhitepiece + anyblackpiece
        record_selector = finder.db.record_selector
        rays = constants.RAYS
        empty_square_games = self.empty_square_games
        piece_square_games = self.piece_square_games
        recordset_cache = self.recordset_cache
        start = self.raycomponents[0]
        final = self.raycomponents[-1]
        for s in start:
            start_square = s[-3:-1]
            rs = rays[start_square]
            for f in final:
                final_square = f[-3:-1]
                if final_square not in rs:
                    continue
                for ps in (
                    s,
                    f,
                ):
                    if ps not in piece_square_games:
                        if ps[-1] in anypiece:
                            w = record_selector(
                                " ".join(
                                    (
                                        filespec.SQUAREMOVE_FIELD_DEF,
                                        "eq",
                                        ps,
                                    )
                                )
                            )
                        else:
                            w = record_selector(
                                " ".join(
                                    (
                                        filespec.PIECESQUAREMOVE_FIELD_DEF,
                                        "eq",
                                        ps,
                                    )
                                )
                            )
                        w.lex()
                        w.parse()
                        w.evaluate(finder)
                        piece_square_games.add(ps)
                        recordset_cache[ps] = w.node.result.answer
                i = start_square, final_square
                self._add_recordset_to_ray_games(
                    recordset_cache[s], recordset_cache[f], i, finder
                )
        start = self.emptycomponents[0]
        final = self.emptycomponents[-1]
        for s in start:
            start_square = s[-2:]
            rs = rays[start_square]
            for f in final:
                final_square = f[-2:]
                if final_square not in rs:
                    continue
                for ps in (
                    s,
                    f,
                ):
                    if ps not in empty_square_games:
                        w = record_selector(
                            " ".join(
                                (
                                    "not",
                                    "(",
                                    filespec.SQUAREMOVE_FIELD_DEF,
                                    "eq",
                                    ps + anywhitepiece,
                                    "or",
                                    ps + anyblackpiece,
                                    ")",
                                )
                            )
                        )
                        w.lex()
                        w.parse()
                        w.evaluate(finder)
                        empty_square_games.add(ps)
                        recordset_cache[ps] = w.node.result.answer
                i = start_square, final_square
                self._add_recordset_to_ray_games(
                    recordset_cache[s], recordset_cache[f], i, finder
                )
        start = self.raycomponents[0]
        final = self.raycomponents[-1]
        for e in self.emptycomponents[0]:
            if e not in empty_square_games:
                continue
            start_square = e[-2:]
            sc = recordset_cache[e]
            for f in final:
                fc = recordset_cache[f]
                final_square = f[-3:-1]
                i = start_square, final_square
                self._add_recordset_to_ray_games(sc, fc, i, finder)
        for e in self.emptycomponents[-1]:
            if e not in empty_square_games:
                continue
            final_square = e[-2:]
            fc = recordset_cache[e]
            for s in start:
                sc = recordset_cache[s]
                start_square = s[-3:-1]
                i = start_square, final_square
                self._add_recordset_to_ray_games(sc, fc, i, finder)

    def _add_recordset_to_ray_games(self, start, final, rayindex, finder):
        """Remove ray-end squares with no game references"""
        rg = start & final
        if rg.count_records():
            if rayindex in self.ray_games:
                self.ray_games[rayindex] |= rg
            else:
                self.ray_games[rayindex] = rg

    def find_games_for_middle_squares(self, finder):
        """Remove ray-end squares with no game references"""
        anywhitepiece = constants.ANY_WHITE_PIECE_NAME
        anyblackpiece = constants.ANY_BLACK_PIECE_NAME
        anypiece = anywhitepiece + anyblackpiece
        nopiece = constants.EMPTY_SQUARE_NAME
        record_selector = finder.db.record_selector
        internal_ray_length = len(self.raycomponents) - 2
        empty_square_games = self.empty_square_games
        piece_square_games = self.piece_square_games
        recordset_cache = self.recordset_cache
        raycomponents = self.raycomponents
        internal_raycomponents = raycomponents[1:-1]
        mvi = move_number_str(self.move_number) + self.variation_code
        c_sqi = [{}]  # Maybe the empty square index values?
        for e, rc in enumerate(internal_raycomponents):
            sqi = {}
            c_sqi.append(sqi)
            for i in rc:
                sqi.setdefault(i[-3:-1], set()).add(i)
        for start, final in self.ray_games:
            line = constants.RAYS[start][final][1:-1]
            if len(line) < len(internal_raycomponents):
                continue
            mapraytoline = MAP_RAY_TO_LINE[len(internal_raycomponents)]
            raygames = []
            for mrtl in mapraytoline:
                if len(line) < 6:  # mapraytoline[7] = 0 avoids this test.
                    if mrtl[len(line)]:
                        break
                linesets = []
                for e, v in enumerate(mrtl[: len(line)]):
                    if line[e] not in c_sqi[v]:
                        if v:
                            linesets.clear()
                            break
                        c_sqi[v][line[e]] = {mvi + line[e] + nopiece}
                    linesets.append(c_sqi[v][line[e]])
                linegames = []
                for lg in linesets:
                    squareset = finder.db.recordlist_nil(finder.dbset)
                    linegames.append(squareset)
                    for i in lg:
                        if i in recordset_cache:
                            squareset |= recordset_cache[i]
                            continue
                        if i[-1] == nopiece:
                            w = record_selector(
                                " ".join(
                                    (
                                        "not",
                                        "(",
                                        filespec.SQUAREMOVE_FIELD_DEF,
                                        "eq",
                                        i[:-1] + anywhitepiece,
                                        "or",
                                        i[:-1] + anyblackpiece,
                                        ")",
                                    )
                                )
                            )
                            w.lex()
                            w.parse()
                            w.evaluate(finder)
                            recordset_cache[i] = w.node.result.answer
                            squareset |= recordset_cache[i]
                            continue
                        if i[-1] in anypiece:
                            w = record_selector(
                                " ".join(
                                    (
                                        filespec.SQUAREMOVE_FIELD_DEF,
                                        "eq",
                                        i,
                                    )
                                )
                            )
                        else:
                            w = record_selector(
                                " ".join(
                                    (
                                        filespec.PIECESQUAREMOVE_FIELD_DEF,
                                        "eq",
                                        i,
                                    )
                                )
                            )
                        w.lex()
                        w.parse()
                        w.evaluate(finder)
                        recordset_cache[i] = w.node.result.answer
                        squareset |= recordset_cache[i]
                if linegames:
                    squareset = linegames.pop() & self.ray_games[index]
                    for lg in linegames:
                        squareset &= lg
                    raygames.append(squareset)
            if raygames:
                rayset = raygames.pop()
                for rg in raygames:
                    rayset |= rg
                self.ray_games[start, final].replace_records(rayset)
            else:
                self.ray_games[start, final].replace_records(
                    finder.db.recordlist_nil(finder.dbset)
                )


# This function belong in, and has been moved to, a chesstab module.  It came
# from chessql.core.piecedesignator.PieceDesignator, it was a staticmethod, but
# chesstab has no subclass of of PieceDesignator (yet), and rayfilter is only
# user at present.
# The alternative definitions are retained at present.
def piece_square_to_index(designator_set, index_prefix):
    """Convert piece designator set values to index format: Qa4 to a4Q.

    Assumed that having all index values for a square adjacent is better
    than having index values for piece together, despite the need for
    conversion.

    """
    ecs = piecedesignator.PieceDesignator._expand_composite_square
    ds = set()
    for ps in designator_set:
        if len(ps) != 1:
            ds.add(index_prefix + ps)
        else:
            ds.update(
                {
                    index_prefix + ps + s
                    for s in ecs(
                        FILE_NAMES[0],
                        FILE_NAMES[-1],
                        RANK_NAMES[0],
                        RANK_NAMES[-1],
                    )
                }
            )
    return ds


# If 'square piece' is better order than 'piece square'
def piece_square_to_index(designator_set, index_prefix):
    """Convert piece designator set values to index format: Qa4 to a4Q.

    Assumed that having all index values for a square adjacent is better
    than having index values for piece together, despite the need for
    conversion.

    """
    ecs = piecedesignator.PieceDesignator._expand_composite_square
    ds = set()
    for ps in designator_set:
        if len(ps) != 1:
            ds.add(index_prefix + ps[1:] + ps[0])
        else:
            ds.update(
                {
                    index_prefix + s + ps
                    for s in ecs(
                        FILE_NAMES[0],
                        FILE_NAMES[-1],
                        RANK_NAMES[0],
                        RANK_NAMES[-1],
                    )
                }
            )
    return ds


def move_number_str(move_number):
    """Return hex(move_number) values without leading '0x' but prefixed with
    string length.
    """
    # Adapted from module pgn_read.core.parser method add_move_to_game().
    try:
        return MOVE_NUMBER_KEYS[move_number]
    except IndexError:
        c = hex(move_number)
        return str(len(c) - 2) + c[2:]



# board.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Classes to draw positions on chess boards.

The Board class is used to show positions in a game of chess.

The PartialBoard class is used to show partial positions representing criteria
for selecting games by pattern of pieces on the board.

Fonts such as Chess Merida, if installed, are used to represent pieces on the
board.

Character equivalents for the pieces if none of these fonts are installed:

l  King
w  Queen
t  Rook
v  Bishop
m  Knight
o  Pawn

PartialBoard uses additional characters:

-  no piece
?  any piece
X  any white piece
x  any black piece
   space means any or no piece: the square is ignored when selecting games.
"""

import tkinter
import tkinter.font

from solentware_misc.gui.exceptionhandler import ExceptionHandler

from pgn_read.core.constants import (
    FEN_WHITE_KING,
    FEN_WHITE_QUEEN,
    FEN_WHITE_ROOK,
    FEN_WHITE_BISHOP,
    FEN_WHITE_KNIGHT,
    FEN_WHITE_PAWN,
    FEN_BLACK_KING,
    FEN_BLACK_QUEEN,
    FEN_BLACK_ROOK,
    FEN_BLACK_BISHOP,
    FEN_BLACK_KNIGHT,
    FEN_BLACK_PAWN,
    FEN_WHITE_PIECES,
    FEN_BLACK_PIECES,
)
from pgn_read.core.squares import Squares

from . import constants
from ..core.constants import NOPIECE

# Characters are black pieces on light square in the four Chess fonts, Cases
# Lucena Merida and Motif, by Armando H Marroquin.
# Fonts were downloaded from www.enpassant.dk/chess/fonteng.htm
_pieces = {
    NOPIECE: "",
    FEN_WHITE_KING: "l",
    FEN_WHITE_QUEEN: "w",
    FEN_WHITE_ROOK: "t",
    FEN_WHITE_BISHOP: "v",
    FEN_WHITE_KNIGHT: "m",
    FEN_WHITE_PAWN: "o",
    FEN_BLACK_KING: "l",
    FEN_BLACK_QUEEN: "w",
    FEN_BLACK_ROOK: "t",
    FEN_BLACK_BISHOP: "v",
    FEN_BLACK_KNIGHT: "m",
    FEN_BLACK_PAWN: "o",
}


class Board(ExceptionHandler):

    """Chess board widget.

    Frame containing an 8x8 grid of Text widgets representing chess board
    with a font used to denote the pieces.

    """

    litecolor = constants.LITECOLOR
    darkcolor = constants.DARKCOLOR
    whitecolor = constants.WHITECOLOR
    blackcolor = constants.BLACKCOLOR
    boardfont = constants.PIECES_ON_BOARD_FONT

    def __init__(self, master, boardborder=2, boardfont=None, ui=None):
        """Create board widget.

        The container catches application resizing and reconfigures
        Board to it's new size. The board then processes the Canvases to
        adjust fonts. Neither propagates geometry changes to it's master.

        """
        self.ui = ui
        if boardfont:
            self.boardfont = boardfont

        self.squares = dict()
        self.boardsquares = dict()
        # self.boardfont is name of named font or a font instance
        try:
            self.font = tkinter.font.nametofont(self.boardfont).copy()
        except (AttributeError, tkinter.TclError):
            self.font = self.boardfont.copy()
        self.container = tkinter.Frame(master, width=0, height=0)
        self.container.bind(
            "<Configure>", self.try_event(self.on_configure_container)
        )
        self.board = tkinter.Frame(
            self.container, borderwidth=boardborder, relief=tkinter.SUNKEN
        )
        self.board.pack(anchor=tkinter.W)
        self.board.grid_propagate(False)
        for i in range(8):
            self.board.grid_rowconfigure(i, weight=1, uniform="r")
            self.board.grid_columnconfigure(i, weight=1, uniform="c")
            for j in range(8):
                s = i * 8 + j
                if (i + j) % 2 == 0:
                    scolor = self.litecolor
                else:
                    scolor = self.darkcolor
                self.boardsquares[s] = t = tkinter.Label(
                    self.board, font=self.font, background=scolor
                )
                t.grid(column=j, row=i, sticky=tkinter.NSEW)

    def configure_font(self, side):
        """Adjust font size after container widget has been resized."""
        self.font.configure(size=-(side * 3) // 32)

    def on_configure_container(self, event=None):
        """Reconfigure board after container widget has been resized."""
        cw = self.container.winfo_width()
        ch = self.container.winfo_height()
        side = min(cw, ch)
        self.board.configure(width=side, height=side)
        self.configure_font(side)
        self.draw_board()

    def draw_board(self):
        """Set font size to match board size and redraw pieces."""
        for i in self.squares:
            p = self.squares[i]
            if p in FEN_WHITE_PIECES:
                pcolor = self.whitecolor
            elif p in FEN_BLACK_PIECES:
                pcolor = self.blackcolor
            elif p == NOPIECE:
                if i % 2 == 0:
                    pcolor = self.darkcolor
                else:
                    pcolor = self.litecolor
            else:
                continue
            self.boardsquares[i].configure(foreground=pcolor, text=_pieces[p])

    def get_top_widget(self):
        """Return top level frame of this widget."""
        return self.container

    def set_color_scheme(self):
        """Set background color for Canvas for each square."""
        for i in range(8):
            for j in range(8):
                s = i * 8 + j
                if (i + j) % 2 == 0:
                    scolor = self.darkcolor
                else:
                    scolor = self.litecolor
                self.boardsquares[s].configure(background=scolor)

    def set_board(self, board):
        """Redraw widget to display the new position in board.

        board is a list of pieces where element index maps to square.

        """
        sq = self.squares
        occupied = list(sq.keys())
        sq.clear()
        squares = Squares.squares
        for s, p in board.items():
            sq[squares[s].number] = p.name
        for s in occupied:
            if s not in sq:
                sq[s] = NOPIECE
        self.draw_board()
        for s in occupied:
            if sq[s] == NOPIECE:
                del sq[s]


class PartialBoard(Board):

    """Partial board widget.

    Customise Board to display wildpieces.

    """

    wildpiecesfont = constants.WILDPIECES_ON_BOARD_FONT

    def __init__(self, master, **ka):
        """Create partial board widget.

        Define the font for wildpieces then delegate to superclass

        """
        # self.wildpiecesfont is name of named font or a font instance
        try:
            self.wildfont = tkinter.font.nametofont(self.wildpiecesfont).copy()
        except AttributeError:
            self.wildfont = self.wildpiecesfont.copy()
        super(PartialBoard, self).__init__(master, **ka)

    def configure_font(self, side):
        """Adjust font size after container widget has been resized."""
        self.wildfont.configure(size=-(side * 3) // 32)
        super(PartialBoard, self).configure_font(side)

    def draw_board(self):
        """Set font size to match board size and redraw pieces."""

        # Obsolescent comment because board no longer shown for partial
        # position after conversion to CQL statements.

        # Explicit setting of fonts used to get Microsoft Windows to show
        # wildpiece characters on board.
        # Possibly good fortune that wildpiece characters are shown without
        # explicit action on FreeBSD.
        # As implemented the font size used does not vary with square size.
        # Perhaps try adjust font size as well but wildpiece characters may
        # look better if smaller then pieces.

        # End obsolescent comment.

        for i in self.squares:
            font = self.font
            p = self.squares[i]
            if p in FEN_WHITE_PIECES:
                pcolor = self.whitecolor
            elif p in FEN_BLACK_PIECES:
                pcolor = self.blackcolor
            elif p == NOPIECE:
                if i % 2 == 0:
                    pcolor = self.darkcolor
                else:
                    pcolor = self.litecolor
            else:
                continue
            self.boardsquares[i].configure(
                font=font, foreground=pcolor, text=_pieces.get(p, " ")
            )



# exporters.py
# Copyright 2013 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess game, repertoire, and partial position exporters.
"""

import os

from pgn_read.core.parser import PGN

from . import chessrecord, filespec
from .cqlstatement import CQLStatement


def export_all_games_text(database, filename):
    """Export games in database to text file in internal record format."""
    if filename is None:
        return True
    rr = chessrecord.ChessDBrecordGameText()
    rr.set_database(database)
    gamesout = open(filename, "w", encoding="iso-8859-1")
    cursor = database.database_cursor(
        filespec.GAMES_FILE_DEF, filespec.GAMES_FILE_DEF
    )
    try:
        r = cursor.first()
        while r:
            rr.load_record(r)
            gamesout.write(rr.get_srvalue())
            gamesout.write("\n")
            r = cursor.next()
    finally:
        cursor.close()
        gamesout.close()
    return True


def export_all_games_pgn(database, filename):
    """Export all database games in PGN export format."""
    if filename is None:
        return True
    rr = chessrecord.ChessDBrecordGame()
    rr.set_database(database)
    all_games_output = None
    no_games_output = True
    games_for_date = []
    prev_date = None
    gamesout = open(filename, "w")
    cursor = database.database_cursor(
        filespec.GAMES_FILE_DEF, filespec.PGN_DATE_FIELD_DEF
    )
    try:
        r = cursor.first()
        while r:
            if r[0] != prev_date:
                for gfd in sorted(games_for_date):
                    gamesout.write(gfd[0])
                    gamesout.write("\n")
                    gamesout.write(gfd[2])
                    gamesout.write("\n")
                    gamesout.write(gfd[1])
                    gamesout.write("\n\n")
                prev_date = r[0]
                games_for_date = []
            g = database.get_primary_record(filespec.GAMES_FILE_DEF, r[1])
            try:
                rr.load_record(g)
            except StopIteration:
                break
            if rr.value.collected_game.is_pgn_valid_export_format():
                games_for_date.append(
                    rr.value.collected_game.get_export_pgn_elements()
                )
                if all_games_output is None:
                    all_games_output = True
                    no_games_output = False
            elif all_games_output:
                if not no_games_output:
                    all_games_output = False
            r = cursor.next()
        for gfd in sorted(games_for_date):
            gamesout.write(gfd[0])
            gamesout.write("\n")
            gamesout.write(gfd[2])
            gamesout.write("\n")
            gamesout.write(gfd[1])
            gamesout.write("\n\n")
    finally:
        cursor.close()
        gamesout.close()
    return all_games_output


def export_all_games_pgn_import_format(database, filename):
    """Export all database games in a PGN inport format."""
    if filename is None:
        return True
    rr = chessrecord.ChessDBrecordGame()
    rr.set_database(database)
    all_games_output = None
    no_games_output = True
    gamesout = open(filename, "w")
    cursor = database.database_cursor(
        filespec.GAMES_FILE_DEF, filespec.GAMES_FILE_DEF
    )
    try:
        r = cursor.first()
        while r:
            try:
                rr.load_record(r)
            except StopIteration:
                break
            if rr.value.collected_game.is_pgn_valid_export_format():
                gamesout.write(
                    get_game_pgn_import_format(rr.value.collected_game)
                )
                if all_games_output is None:
                    all_games_output = True
                    no_games_output = False
            elif all_games_output:
                if not no_games_output:
                    all_games_output = False
            r = cursor.next()
    finally:
        cursor.close()
        gamesout.close()
    return all_games_output


def export_all_games_pgn_no_comments(database, filename):
    """Export all database games in PGN export format excluding comments."""
    if filename is None:
        return True
    rr = chessrecord.ChessDBrecordGame()
    rr.set_database(database)
    all_games_output = None
    no_games_output = True
    games_for_date = []
    prev_date = None
    gamesout = open(filename, "w")
    cursor = database.database_cursor(
        filespec.GAMES_FILE_DEF, filespec.PGN_DATE_FIELD_DEF
    )
    try:
        r = cursor.first()
        while r:
            if r[0] != prev_date:
                for gfd in sorted(games_for_date):
                    gamesout.write(gfd[0])
                    gamesout.write("\n")
                    gamesout.write(gfd[2])
                    gamesout.write("\n")
                    gamesout.write(gfd[1])
                    gamesout.write("\n\n")
                prev_date = r[0]
                games_for_date = []
            g = database.get_primary_record(filespec.GAMES_FILE_DEF, r[1])
            try:
                rr.load_record(g)
            except StopIteration:
                break
            if rr.value.collected_game.is_pgn_valid_export_format():
                games_for_date.append(
                    rr.value.collected_game.get_export_pgn_rav_elements()
                )
                if all_games_output is None:
                    all_games_output = True
                    no_games_output = False
            elif all_games_output:
                if not no_games_output:
                    all_games_output = False
            r = cursor.next()
        for gfd in sorted(games_for_date):
            gamesout.write(gfd[0])
            gamesout.write("\n")
            gamesout.write(gfd[2])
            gamesout.write("\n")
            gamesout.write(gfd[1])
            gamesout.write("\n\n")
    finally:
        cursor.close()
        gamesout.close()
    return all_games_output


def export_all_games_pgn_no_comments_no_ravs(database, filename):
    """Export all database games in PGN export format excluding comments and
    RAVs.

    """
    if filename is None:
        return True
    rr = chessrecord.ChessDBrecordGame()
    rr.set_database(database)
    all_games_output = None
    no_games_output = True
    games_for_date = []
    prev_date = None
    gamesout = open(filename, "w")
    cursor = database.database_cursor(
        filespec.GAMES_FILE_DEF, filespec.PGN_DATE_FIELD_DEF
    )
    try:
        r = cursor.first()
        while r:
            if r[0] != prev_date:
                for gfd in sorted(games_for_date):
                    gamesout.write(gfd[0])
                    gamesout.write("\n")
                    gamesout.write(gfd[2])
                    gamesout.write("\n")
                    gamesout.write(gfd[1])
                    gamesout.write("\n\n")
                prev_date = r[0]
                games_for_date = []
            g = database.get_primary_record(filespec.GAMES_FILE_DEF, r[1])
            try:
                rr.load_record(g)
            except StopIteration:
                break
            cg = rr.value.collected_game
            if cg.is_pgn_valid_export_format():
                strt = cg.get_seven_tag_roster_tags()
                nstrt = cg.get_non_seven_tag_roster_tags()
                am = cg.get_archive_movetext()
                games_for_date.append((strt, am, nstrt))
                if all_games_output is None:
                    all_games_output = True
                    no_games_output = False
            elif all_games_output:
                if not no_games_output:
                    all_games_output = False
            r = cursor.next()
        for gfd in sorted(games_for_date):
            gamesout.write(gfd[0])
            gamesout.write("\n")
            gamesout.write(gfd[2])
            gamesout.write("\n")
            gamesout.write(gfd[1])
            gamesout.write("\n\n")
    finally:
        cursor.close()
        gamesout.close()
    return all_games_output


def export_all_games_pgn_reduced_export_format(database, filename):
    """Export all database games in PGN reduced export format."""
    if filename is None:
        return True
    rr = chessrecord.ChessDBrecordGame()
    rr.set_database(database)
    all_games_output = None
    no_games_output = True
    games_for_date = []
    prev_date = None
    gamesout = open(filename, "w")
    cursor = database.database_cursor(
        filespec.GAMES_FILE_DEF, filespec.PGN_DATE_FIELD_DEF
    )
    try:
        r = cursor.first()
        while r:
            if r[0] != prev_date:
                for gfd in sorted(games_for_date):
                    gamesout.write(gfd[0])
                    gamesout.write("\n")
                    gamesout.write(gfd[1])
                    gamesout.write("\n\n")
                prev_date = r[0]
                games_for_date = []
            g = database.get_primary_record(filespec.GAMES_FILE_DEF, r[1])
            try:
                rr.load_record(g)
            except StopIteration:
                break
            if rr.value.collected_game.is_pgn_valid_export_format():
                games_for_date.append(
                    rr.value.collected_game.get_archive_pgn_elements()
                )
                if all_games_output is None:
                    all_games_output = True
                    no_games_output = False
            elif all_games_output:
                if not no_games_output:
                    all_games_output = False
            r = cursor.next()
        for gfd in sorted(games_for_date):
            gamesout.write(gfd[0])
            gamesout.write("\n")
            gamesout.write(gfd[1])
            gamesout.write("\n\n")
    finally:
        cursor.close()
        gamesout.close()
    return all_games_output


def export_all_repertoires_pgn(database, filename):
    """Export all repertoires in PGN export format."""
    if filename is None:
        return True
    rr = chessrecord.ChessDBrecordRepertoire()
    rr.set_database(database)
    gamesout = open(filename, "w")
    cursor = database.database_cursor(
        filespec.REPERTOIRE_FILE_DEF, filespec.REPERTOIRE_FILE_DEF
    )
    try:
        r = cursor.first()
        while r:
            rr.load_record(r)
            if rr.value.collected_game.is_pgn_valid():
                gamesout.write(rr.value.collected_game.get_repertoire_pgn())
                gamesout.write("\n\n")
            r = cursor.next()
    finally:
        cursor.close()
        gamesout.close()
    return True


def export_all_repertoires_pgn_no_comments(database, filename):
    """Export all repertoires in PGN export format without comments."""
    if filename is None:
        return True
    rr = chessrecord.ChessDBrecordRepertoire()
    rr.set_database(database)
    gamesout = open(filename, "w")
    cursor = database.database_cursor(
        filespec.REPERTOIRE_FILE_DEF, filespec.REPERTOIRE_FILE_DEF
    )
    try:
        r = cursor.first()
        while r:
            rr.load_record(r)
            if rr.value.collected_game.is_pgn_valid():
                gamesout.write(
                    rr.value.collected_game.get_repertoire_pgn_no_comments()
                )
                gamesout.write("\n\n")
            r = cursor.next()
    finally:
        cursor.close()
        gamesout.close()
    return True


def export_all_repertoires_pgn_import_format(database, filename):
    """Export all repertoires in a PGN import format."""
    if filename is None:
        return True
    rr = chessrecord.ChessDBrecordRepertoire()
    rr.set_database(database)
    gamesout = open(filename, "w")
    cursor = database.database_cursor(
        filespec.REPERTOIRE_FILE_DEF, filespec.REPERTOIRE_FILE_DEF
    )
    try:
        r = cursor.first()
        while r:
            rr.load_record(r)
            if rr.value.collected_game.is_pgn_valid():
                gamesout.write(
                    get_game_pgn_import_format(rr.value.collected_game)
                )
                gamesout.write("\n\n")
            r = cursor.next()
    finally:
        cursor.close()
        gamesout.close()
    return True


def export_all_repertoires_text(database, filename):
    """Export repertoires in database to text file in internal record format."""
    if filename is None:
        return True
    rr = chessrecord.ChessDBrecordGameText()
    rr.set_database(database)
    gamesout = open(filename, "w")
    cursor = database.database_cursor(
        filespec.REPERTOIRE_FILE_DEF, filespec.REPERTOIRE_FILE_DEF
    )
    try:
        r = cursor.first()
        while r:
            rr.load_record(r)
            gamesout.write(rr.get_srvalue())
            gamesout.write("\n")
            r = cursor.next()
    finally:
        cursor.close()
        gamesout.close()
    return True


def export_all_positions(database, filename):
    """Export CQL statements in database to text file in internal record
    format.

    """
    if filename is None:
        return True
    rr = chessrecord.ChessDBrecordPartial()
    rr.set_database(database)
    gamesout = open(filename, "w")
    cursor = database.database_cursor(
        filespec.PARTIAL_FILE_DEF, filespec.PARTIAL_FILE_DEF
    )
    try:
        r = cursor.first()
        while r:
            rr.load_record(r)
            gamesout.write(rr.get_srvalue())
            gamesout.write("\n")
            r = cursor.next()
    finally:
        cursor.close()
        gamesout.close()
    return True


def export_selected_games_pgn_import_format(grid, filename):
    """Export selected records in a PGN import format.

    If any records are bookmarked just the bookmarked records are exported,
    otherwise all records selected for display in the grid are exported.

    """
    if filename is None:
        return True
    database = grid.get_data_source().dbhome
    primary = database.is_primary(
        grid.get_data_source().dbset, grid.get_data_source().dbname
    )
    rr = chessrecord.ChessDBrecordGame()
    rr.set_database(database)
    games = []
    all_games_output = True
    if grid.bookmarks:
        for b in grid.bookmarks:
            rr.load_record(
                database.get_primary_record(
                    filespec.GAMES_FILE_DEF, b[0 if primary else 1]
                )
            )
            if rr.value.collected_game.is_pgn_valid_export_format():
                games.append(
                    get_game_pgn_import_format(rr.value.collected_game)
                )
            else:
                all_games_output = False
    elif grid.partial:
        cursor = grid.get_cursor()
        try:
            if primary:
                r = cursor.first()
            else:
                r = cursor.nearest(
                    database.encode_record_selector(grid.partial)
                )
            while r:
                if not primary:
                    if not r[0].startswith(grid.partial):
                        break
                rr.load_record(
                    database.get_primary_record(
                        filespec.GAMES_FILE_DEF, r[0 if primary else 1]
                    )
                )
                if rr.value.collected_game.is_pgn_valid_export_format():
                    games.append(
                        get_game_pgn_import_format(rr.value.collected_game)
                    )
                else:
                    all_games_output = False
                r = cursor.next()
        finally:
            cursor.close()
    else:
        cursor = grid.get_cursor()

        # For all grids except ones displayed via 'Select | Rule | List Games'
        # the 'r = cursor.next()' can be immediately after the 'while True:'
        # statement, and the 'r = cursor.first()' statement is redundant.
        # I think this implies a problem in the solentware_base RecordsetCursor
        # classes for each database engine since the 'finally:' clause should
        # kill the cursor.
        # The problem is only the first request outputs all the records to the
        # file.  Subsequent requests find no records to output, except that
        # doing some scrolling action resets the cursor and the next request
        # outputs all the records before the problem repeats.
        # The other methods in this class with this construct are affected too,
        # but this comment is not repeated.
        try:
            r = cursor.first()
            while True:
                if r is None:
                    break
                rr.load_record(
                    database.get_primary_record(
                        filespec.GAMES_FILE_DEF, r[0 if primary else 1]
                    )
                )
                if rr.value.collected_game.is_pgn_valid_export_format():
                    games.append(
                        get_game_pgn_import_format(rr.value.collected_game)
                    )
                else:
                    all_games_output = False
                r = cursor.next()
        finally:
            cursor.close()

    if len(games) == 0:
        return None
    gamesout = open(filename, "w")
    try:
        for g in games:
            gamesout.write(g)
            gamesout.write("\n\n")
    finally:
        gamesout.close()
    return all_games_output


def export_selected_games_pgn(grid, filename):
    """Export selected records in PGN export format.

    If any records are bookmarked just the bookmarked records are exported,
    otherwise all records selected for display in the grid are exported.

    """
    if filename is None:
        return True
    database = grid.get_data_source().dbhome
    primary = database.is_primary(
        grid.get_data_source().dbset, grid.get_data_source().dbname
    )
    rr = chessrecord.ChessDBrecordGame()
    rr.set_database(database)
    games = []
    all_games_output = True
    if grid.bookmarks:
        for b in grid.bookmarks:
            rr.load_record(
                database.get_primary_record(
                    filespec.GAMES_FILE_DEF, b[0 if primary else 1]
                )
            )
            if rr.value.collected_game.is_pgn_valid_export_format():
                games.append(rr.value.collected_game.get_export_pgn_elements())
            else:
                all_games_output = False
    elif grid.partial:
        cursor = grid.get_cursor()
        try:
            if primary:
                r = cursor.first()
            else:
                r = cursor.nearest(
                    database.encode_record_selector(grid.partial)
                )
            while r:
                if not primary:
                    if not r[0].startswith(grid.partial):
                        break
                rr.load_record(
                    database.get_primary_record(
                        filespec.GAMES_FILE_DEF, r[0 if primary else 1]
                    )
                )
                if rr.value.collected_game.is_pgn_valid_export_format():
                    games.append(
                        rr.value.collected_game.get_export_pgn_elements()
                    )
                else:
                    all_games_output = False
                r = cursor.next()
        finally:
            cursor.close()
    else:
        cursor = grid.get_cursor()

        # For all grids except ones displayed via 'Select | Rule | List Games'
        # the 'r = cursor.next()' can be immediately after the 'while True:'
        # statement, and the 'r = cursor.first()' statement is redundant.
        # I think this implies a problem in the solentware_base RecordsetCursor
        # classes for each database engine since the 'finally:' clause should
        # kill the cursor.
        # The problem is only the first request outputs all the records to the
        # file.  Subsequent requests find no records to output, except that
        # doing some scrolling action resets the cursor and the next request
        # outputs all the records before the problem repeats.
        # The other methods in this class with this construct are affected too,
        # but this comment is not repeated.
        try:
            r = cursor.first()
            while True:
                if r is None:
                    break
                rr.load_record(
                    database.get_primary_record(
                        filespec.GAMES_FILE_DEF, r[0 if primary else 1]
                    )
                )
                if rr.value.collected_game.is_pgn_valid_export_format():
                    games.append(
                        rr.value.collected_game.get_export_pgn_elements()
                    )
                else:
                    all_games_output = False
                r = cursor.next()
        finally:
            cursor.close()

    if len(games) == 0:
        return None
    gamesout = open(filename, "w")
    try:
        for g in sorted(games):
            gamesout.write(g[0])
            gamesout.write("\n")
            gamesout.write(g[2])
            gamesout.write("\n")
            gamesout.write(g[1])
            gamesout.write("\n\n")
    finally:
        gamesout.close()
    return all_games_output


def export_selected_games_pgn_no_comments(grid, filename):
    """Export selected records in PGN export format excluding comments.

    If any records are bookmarked just the bookmarked records are exported,
    otherwise all records selected for display in the grid are exported.

    """
    if filename is None:
        return True
    database = grid.get_data_source().dbhome
    primary = database.is_primary(
        grid.get_data_source().dbset, grid.get_data_source().dbname
    )
    rr = chessrecord.ChessDBrecordGame()
    rr.set_database(database)
    games = []
    all_games_output = True
    if grid.bookmarks:
        for b in grid.bookmarks:
            rr.load_record(
                database.get_primary_record(
                    filespec.GAMES_FILE_DEF, b[0 if primary else 1]
                )
            )
            if rr.value.collected_game.is_pgn_valid_export_format():
                games.append(
                    rr.value.collected_game.get_export_pgn_rav_elements()
                )
            else:
                all_games_output = False
    elif grid.partial:
        cursor = grid.get_cursor()
        try:
            if primary:
                r = cursor.first()
            else:
                r = cursor.nearest(
                    database.encode_record_selector(grid.partial)
                )
            while r:
                if not primary:
                    if not r[0].startswith(grid.partial):
                        break
                rr.load_record(
                    database.get_primary_record(
                        filespec.GAMES_FILE_DEF, r[0 if primary else 1]
                    )
                )
                if rr.value.collected_game.is_pgn_valid_export_format():
                    games.append(
                        rr.value.collected_game.get_export_pgn_rav_elements()
                    )
                else:
                    all_games_output = False
                r = cursor.next()
        finally:
            cursor.close()
    else:
        cursor = grid.get_cursor()
        try:
            r = cursor.first()
            while True:
                if r is None:
                    break
                rr.load_record(
                    database.get_primary_record(
                        filespec.GAMES_FILE_DEF, r[0 if primary else 1]
                    )
                )
                if rr.value.collected_game.is_pgn_valid_export_format():
                    games.append(
                        rr.value.collected_game.get_export_pgn_rav_elements()
                    )
                else:
                    all_games_output = False
                r = cursor.next()
        finally:
            cursor.close()
    if len(games) == 0:
        return None
    gamesout = open(filename, "w")
    try:
        for g in sorted(games):
            gamesout.write(g[0])
            gamesout.write("\n")
            gamesout.write(g[2])
            gamesout.write("\n")
            gamesout.write(g[1])
            gamesout.write("\n\n")
    finally:
        gamesout.close()
    return all_games_output


def export_selected_games_pgn_no_comments_no_ravs(grid, filename):
    """Export selected records in PGN export format excluding comments.

    If any records are bookmarked just the bookmarked records are exported,
    otherwise all records selected for display in the grid are exported.

    """
    if filename is None:
        return True
    database = grid.get_data_source().dbhome
    primary = database.is_primary(
        grid.get_data_source().dbset, grid.get_data_source().dbname
    )
    rr = chessrecord.ChessDBrecordGame()
    rr.set_database(database)
    games = []
    all_games_output = True
    if grid.bookmarks:
        for b in grid.bookmarks:
            rr.load_record(
                database.get_primary_record(
                    filespec.GAMES_FILE_DEF, b[0 if primary else 1]
                )
            )
            cg = rr.value.collected_game
            if cg.is_pgn_valid_export_format():
                strt = cg.get_seven_tag_roster_tags()
                nstrt = cg.get_non_seven_tag_roster_tags()
                am = cg.get_archive_movetext()
                games.append((strt, am, nstrt))
            else:
                all_games_output = False
    elif grid.partial:
        cursor = grid.get_cursor()
        try:
            if primary:
                r = cursor.first()
            else:
                r = cursor.nearest(
                    database.encode_record_selector(grid.partial)
                )
            while r:
                if not primary:
                    if not r[0].startswith(grid.partial):
                        break
                rr.load_record(
                    database.get_primary_record(
                        filespec.GAMES_FILE_DEF, r[0 if primary else 1]
                    )
                )
                cg = rr.value.collected_game
                if cg.is_pgn_valid_export_format():
                    strt = cg.get_seven_tag_roster_tags()
                    nstrt = cg.get_non_seven_tag_roster_tags()
                    am = cg.get_archive_movetext()
                    games.append((strt, am, nstrt))
                else:
                    all_games_output = False
                r = cursor.next()
        finally:
            cursor.close()
    else:
        cursor = grid.get_cursor()
        try:
            r = cursor.first()
            while True:
                if r is None:
                    break
                rr.load_record(
                    database.get_primary_record(
                        filespec.GAMES_FILE_DEF, r[0 if primary else 1]
                    )
                )
                cg = rr.value.collected_game
                if cg.is_pgn_valid_export_format():
                    strt = cg.get_seven_tag_roster_tags()
                    nstrt = cg.get_non_seven_tag_roster_tags()
                    am = cg.get_archive_movetext()
                    games.append((strt, am, nstrt))
                else:
                    all_games_output = False
                r = cursor.next()
        finally:
            cursor.close()
    if len(games) == 0:
        return None
    gamesout = open(filename, "w")
    try:
        for g in sorted(games):
            gamesout.write(g[0])
            gamesout.write("\n")
            gamesout.write(g[2])
            gamesout.write("\n")
            gamesout.write(g[1])
            gamesout.write("\n\n")
    finally:
        gamesout.close()
    return all_games_output


def export_selected_games_pgn_reduced_export_format(grid, filename):
    """Export selected records in grid to PGN file in reduced export format.

    If any records are bookmarked just the bookmarked records are exported,
    otherwise all records selected for display in the grid are exported.

    """
    if filename is None:
        return True
    database = grid.get_data_source().dbhome
    primary = database.is_primary(
        grid.get_data_source().dbset, grid.get_data_source().dbname
    )
    rr = chessrecord.ChessDBrecordGame()
    rr.set_database(database)
    games = []
    all_games_output = True
    if grid.bookmarks:
        for b in grid.bookmarks:
            rr.load_record(
                database.get_primary_record(
                    filespec.GAMES_FILE_DEF, b[0 if primary else 1]
                )
            )
            if rr.value.collected_game.is_pgn_valid_export_format():
                games.append(
                    rr.value.collected_game.get_archive_pgn_elements()
                )
            else:
                all_games_output = False
    elif grid.partial:
        cursor = grid.get_cursor()
        try:
            if primary:
                r = cursor.first()
            else:
                r = cursor.nearest(
                    database.encode_record_selector(grid.partial)
                )
            while r:
                if not primary:
                    if not r[0].startswith(grid.partial):
                        break
                rr.load_record(
                    database.get_primary_record(
                        filespec.GAMES_FILE_DEF, r[0 if primary else 1]
                    )
                )
                if rr.value.collected_game.is_pgn_valid_export_format():
                    games.append(
                        rr.value.collected_game.get_archive_pgn_elements()
                    )
                else:
                    all_games_output = False
                r = cursor.next()
        finally:
            cursor.close()
    else:
        cursor = grid.get_cursor()
        try:
            r = cursor.first()
            while True:
                if r is None:
                    break
                rr.load_record(
                    database.get_primary_record(
                        filespec.GAMES_FILE_DEF, r[0 if primary else 1]
                    )
                )
                if rr.value.collected_game.is_pgn_valid_export_format():
                    games.append(
                        rr.value.collected_game.get_archive_pgn_elements()
                    )
                else:
                    all_games_output = False
                r = cursor.next()
        finally:
            cursor.close()
    if len(games) == 0:
        return None
    gamesout = open(filename, "w")
    try:
        for g in sorted(games):
            gamesout.write(g[0])
            gamesout.write("\n")
            gamesout.write(g[1])
            gamesout.write("\n\n")
    finally:
        gamesout.close()
    return all_games_output


def export_selected_games_text(grid, filename):
    """Export selected records in grid to text file in internal record format.

    If any records are bookmarked just the bookmarked records are exported,
    otherwise all records selected for display in the grid are exported.

    """
    if filename is None:
        return True
    database = grid.get_data_source().dbhome
    primary = database.is_primary(
        grid.get_data_source().dbset, grid.get_data_source().dbname
    )
    rr = chessrecord.ChessDBrecordGame()
    rr.set_database(database)
    games = []
    if grid.bookmarks:
        for b in grid.bookmarks:
            rr.load_record(
                database.get_primary_record(
                    filespec.GAMES_FILE_DEF, b[0 if primary else 1]
                )
            )
            games.append(rr.get_srvalue())
    elif grid.partial:
        cursor = grid.get_cursor()
        try:
            if primary:
                r = cursor.first()
            else:
                r = cursor.nearest(
                    database.encode_record_selector(grid.partial)
                )
            while r:
                if not primary:
                    if not r[0].startswith(grid.partial):
                        break
                rr.load_record(
                    database.get_primary_record(
                        filespec.GAMES_FILE_DEF, r[0 if primary else 1]
                    )
                )
                games.append(rr.get_srvalue())
                r = cursor.next()
        finally:
            cursor.close()
    else:
        cursor = grid.get_cursor()
        try:
            r = cursor.first()
            while True:
                if r is None:
                    break
                rr.load_record(
                    database.get_primary_record(
                        filespec.GAMES_FILE_DEF, r[0 if primary else 1]
                    )
                )
                games.append(rr.get_srvalue())
                r = cursor.next()
        finally:
            cursor.close()
    if len(games) == 0:
        return None
    gamesout = open(filename, "w")
    try:
        for g in games:
            gamesout.write(g)
            gamesout.write("\n")
    finally:
        gamesout.close()
    return True


def export_selected_repertoires_pgn(grid, filename):
    """Export selected repertoires in PGN export format.

    If any records are bookmarked just the bookmarked records are exported,
    otherwise all records selected for display in the grid are exported.

    """
    if filename is None:
        return True
    if grid.bookmarks:
        database = grid.get_data_source().dbhome
        rr = chessrecord.ChessDBrecordRepertoire()
        rr.set_database(database)
        gamesout = open(filename, "w")
        try:
            for b in sorted(grid.bookmarks):
                rr.load_record(
                    database.get_primary_record(
                        filespec.REPERTOIRE_FILE_DEF, b[0]
                    )
                )
                if rr.value.collected_game.is_pgn_valid_export_format():
                    gamesout.write(
                        rr.value.collected_game.get_repertoire_pgn()
                    )
            gamesout = open(filename, "w")
        finally:
            gamesout.close()
        return True
    else:
        export_all_repertoires_pgn(grid.get_data_source().dbhome, filename)
        return True


def export_selected_repertoires_pgn_no_comments(grid, filename):
    """Export selected repertoires in PGN export format without comments.

    If any records are bookmarked just the bookmarked records are exported,
    otherwise all records selected for display in the grid are exported.

    """
    if filename is None:
        return
    if grid.bookmarks:
        database = grid.get_data_source().dbhome
        rr = chessrecord.ChessDBrecordRepertoire()
        rr.set_database(database)
        gamesout = open(filename, "w")
        try:
            for b in sorted(grid.bookmarks):
                rr.load_record(
                    database.get_primary_record(
                        filespec.REPERTOIRE_FILE_DEF, b[0]
                    )
                )
                if rr.value.collected_game.is_pgn_valid_export_format():
                    gamesout.write(
                        rr.value.collected_game.get_repertoire_pgn_no_comments()
                    )
            gamesout = open(filename, "w")
        finally:
            gamesout.close()
        return
    else:
        export_all_repertoires_pgn_no_comments(
            grid.get_data_source().dbhome, filename
        )
        return


def export_selected_repertoires_pgn_import_format(grid, filename):
    """Export selected repertoires in a PGN import format.

    If any records are bookmarked just the bookmarked records are exported,
    otherwise all records selected for display in the grid are exported.

    """
    if filename is None:
        return True
    database = grid.get_data_source().dbhome
    primary = database.is_primary(
        grid.get_data_source().dbset, grid.get_data_source().dbname
    )
    rr = chessrecord.ChessDBrecordRepertoire()
    rr.set_database(database)
    games = []
    all_games_output = True
    if grid.bookmarks:
        for b in grid.bookmarks:
            rr.load_record(
                database.get_primary_record(
                    filespec.REPERTOIRE_FILE_DEF, b[0 if primary else 1]
                )
            )
            if rr.value.collected_game.is_pgn_valid_export_format():
                games.append(
                    get_game_pgn_import_format(rr.value.collected_game)
                )
            else:
                all_games_output = False
    elif grid.partial:
        cursor = grid.get_cursor()
        try:
            if primary:
                r = cursor.first()
            else:
                r = cursor.nearest(
                    database.encode_record_selector(grid.partial)
                )
            while r:
                if not primary:
                    if not r[0].startswith(grid.partial):
                        break
                rr.load_record(
                    database.get_primary_record(
                        filespec.REPERTOIRE_FILE_DEF, r[0 if primary else 1]
                    )
                )
                if rr.value.collected_game.is_pgn_valid_export_format():
                    games.append(
                        get_game_pgn_import_format(rr.value.collected_game)
                    )
                else:
                    all_games_output = False
                r = cursor.next()
        finally:
            cursor.close()
    else:
        cursor = grid.get_cursor()
        try:
            r = cursor.first()
            while True:
                if r is None:
                    break
                rr.load_record(
                    database.get_primary_record(
                        filespec.REPERTOIRE_FILE_DEF, r[0 if primary else 1]
                    )
                )
                if rr.value.collected_game.is_pgn_valid_export_format():
                    games.append(
                        get_game_pgn_import_format(rr.value.collected_game)
                    )
                else:
                    all_games_output = False
                r = cursor.next()
        finally:
            cursor.close()
    if len(games) == 0:
        return None
    gamesout = open(filename, "w")
    try:
        for g in games:
            gamesout.write(g)
            gamesout.write("\n\n")
    finally:
        gamesout.close()
    return all_games_output


def export_selected_repertoires_text(grid, filename):
    """Export selected repertoires to text file in internal record format.

    If any records are bookmarked just the bookmarked records are exported,
    otherwise all records selected for display in the grid are exported.

    """
    if filename is None:
        return True
    database = grid.get_data_source().dbhome
    primary = database.is_primary(
        grid.get_data_source().dbset, grid.get_data_source().dbname
    )
    rr = chessrecord.ChessDBrecordRepertoire()
    rr.set_database(database)
    games = []
    if grid.bookmarks:
        for b in grid.bookmarks:
            rr.load_record(
                database.get_primary_record(
                    filespec.REPERTOIRE_FILE_DEF, b[0 if primary else 1]
                )
            )
            games.append(rr.get_srvalue())
    elif grid.partial:
        cursor = grid.get_cursor()
        try:
            if primary:
                r = cursor.first()
            else:
                r = cursor.nearest(
                    database.encode_record_selector(grid.partial)
                )
            while r:
                if not primary:
                    if not r[0].startswith(grid.partial):
                        break
                rr.load_record(
                    database.get_primary_record(
                        filespec.REPERTOIRE_FILE_DEF, r[0 if primary else 1]
                    )
                )
                games.append(rr.get_srvalue())
                r = cursor.next()
        finally:
            cursor.close()
    else:
        cursor = grid.get_cursor()
        try:
            r = cursor.first()
            while True:
                if r is None:
                    break
                rr.load_record(
                    database.get_primary_record(
                        filespec.REPERTOIRE_FILE_DEF, r[0 if primary else 1]
                    )
                )
                games.append(rr.get_srvalue())
                r = cursor.next()
        finally:
            cursor.close()
    if len(games) == 0:
        return None
    gamesout = open(filename, "w")
    try:
        for g in games:
            gamesout.write(g)
            gamesout.write("\n")
    finally:
        gamesout.close()
    return True


def export_selected_positions(grid, filename):
    """Export CQL statements in grid to textfile."""
    if filename is None:
        return
    if grid.bookmarks:
        database = grid.get_data_source().dbhome
        rr = chessrecord.ChessDBrecordPartial()
        rr.set_database(database)
        gamesout = open(filename, "w")
        try:
            for b in sorted(grid.bookmarks):
                rr.load_record(
                    database.get_primary_record(
                        filespec.PARTIAL_FILE_DEF, b[0]
                    )
                )
                gamesout.write(rr.get_srvalue())
                gamesout.write("\n")
            gamesout = open(filename, "w")
        finally:
            gamesout.close()
        return
    else:
        database = grid.get_data_source().dbhome
        rr = chessrecord.ChessDBrecordPartial()
        rr.set_database(database)
        gamesout = open(filename, "w")
        cursor = database.database_cursor(
            filespec.PARTIAL_FILE_DEF, filespec.PARTIAL_FILE_DEF
        )
        try:
            r = cursor.first()
            while r:
                rr.load_record(r)
                gamesout.write(rr.get_srvalue())
                gamesout.write("\n")
                r = cursor.next()
        finally:
            cursor.close()
            gamesout.close()
        return


def export_selected_selection_rules(grid, filename):
    """Export selected selection rule statements to textfile."""
    if filename is None:
        return
    if grid.bookmarks:
        database = grid.get_data_source().dbhome
        rr = chessrecord.ChessDBrecordQuery()
        rr.set_database(database)
        gamesout = open(filename, "w")
        try:
            for b in sorted(grid.bookmarks):
                rr.load_record(
                    database.get_primary_record(
                        filespec.SELECTION_FILE_DEF, b[0]
                    )
                )
                gamesout.write(rr.get_srvalue())
                gamesout.write("\n")
            gamesout = open(filename, "w")
        finally:
            gamesout.close()
        return
    else:
        database = grid.get_data_source().dbhome
        rr = chessrecord.ChessDBrecordQuery()
        rr.set_database(database)
        gamesout = open(filename, "w")
        cursor = database.database_cursor(
            filespec.SELECTION_FILE_DEF, filespec.SELECTION_FILE_DEF
        )
        try:
            r = cursor.first()
            while r:
                rr.load_record(r)
                gamesout.write(rr.get_srvalue())
                gamesout.write("\n")
                r = cursor.next()
        finally:
            cursor.close()
            gamesout.close()
        return


def export_single_game_pgn_reduced_export_format(collected_game, filename):
    """Export collected_game to PGN file in reduced export format.

    Caller should test is_pgn_valid_export_format before picking filename.

    """
    if filename is None:
        return
    gamesout = open(filename, "w")
    try:
        gamesout.write(collected_game.get_seven_tag_roster_tags())
        gamesout.write("\n")
        gamesout.write(collected_game.get_archive_movetext())
        gamesout.write("\n\n")
    finally:
        gamesout.close()


def export_single_game_pgn(collected_game, filename):
    """Export collected_game to filename in PGN export format.

    Caller should test is_pgn_valid_export_format before picking filename.

    """
    if filename is None:
        return
    gamesout = open(filename, "w")
    try:
        gamesout.write(collected_game.get_seven_tag_roster_tags())
        gamesout.write("\n")
        gamesout.write(collected_game.get_non_seven_tag_roster_tags())
        gamesout.write("\n")
        gamesout.write(collected_game.get_all_movetext_in_pgn_export_format())
        gamesout.write("\n\n")
    finally:
        gamesout.close()


def export_single_game_pgn_no_comments_no_ravs(collected_game, filename):
    """Export collected_game to filename in PGN export format without comments
    or RAVs.

    Caller should test is_pgn_valid_export_format before picking filename.

    """
    if filename is None:
        return
    gamesout = open(filename, "w")
    try:
        gamesout.write(collected_game.get_seven_tag_roster_tags())
        gamesout.write("\n")
        gamesout.write(collected_game.get_non_seven_tag_roster_tags())
        gamesout.write("\n")
        gamesout.write(collected_game.get_archive_movetext())
        gamesout.write("\n\n")
    finally:
        gamesout.close()
    return True


def export_single_game_pgn_no_comments(collected_game, filename):
    """Export collected_game to filename in PGN export format without comments.

    Caller should test is_pgn_valid_export_format before picking filename.

    """
    if filename is None:
        return
    gamesout = open(filename, "w")
    try:
        gamesout.write(collected_game.get_seven_tag_roster_tags())
        gamesout.write("\n")
        gamesout.write(collected_game.get_non_seven_tag_roster_tags())
        gamesout.write("\n")
        gamesout.write(
            collected_game.get_movetext_without_comments_in_pgn_export_format()
        )
        gamesout.write("\n\n")
    finally:
        gamesout.close()
    return True


def export_single_game_pgn_import_format(collected_game, filename):
    """Export collected_game to pgn file in a PGN import format."""
    if filename is None:
        return
    gamesout = open(filename, "w")
    try:
        gamesout.write(get_game_pgn_import_format(collected_game))
    finally:
        gamesout.close()


def export_single_game_text(collected_game, filename):
    """Export collected_game to text file in internal format."""
    if filename is None:
        return
    internal_format = next(PGN().read_games(collected_game.get_text_of_game()))
    gamesout = open(filename, "w")
    try:
        gamesout.write(internal_format.get_text_of_game())
        gamesout.write("\n")
    finally:
        gamesout.close()


def export_single_repertoire_pgn(collected_game, filename):
    """Export repertoire in import format PGN to *.pgn file.

    Caller should test is_pgn_valid_export_format before picking filename.

    """
    if filename is None:
        return
    gamesout = open(filename, "w")
    try:
        gamesout.write(collected_game.get_repertoire_pgn())
    finally:
        gamesout.close()


def export_single_repertoire_pgn_no_comments(collected_game, filename):
    """Export repertoire in PGN export format without comments.

    Caller should test is_pgn_valid_export_format before picking filename.

    """
    if filename is None:
        return
    gamesout = open(filename, "w")
    try:
        gamesout.write(collected_game.get_repertoire_pgn_no_comments())
    finally:
        gamesout.close()


def export_single_position(partialposition, filename):
    """Export CQL statement to textfile."""
    if filename is None:
        return
    sp = CQLStatement()
    sp.process_statement(partialposition)
    if not sp.is_statement():
        return
    gamesout = open(filename, "w")
    try:
        gamesout.write(sp.get_name_position_text())
    finally:
        gamesout.close()


# Derived from get_all_movetext_in_pgn_export_format method in
# pgn_read.core.game module.
# At 13 Jan 2021 on a sample of two other non-commercial open source chess
# database products both require a restricted line length to allow extraction
# of the game score.  One of them looks like it needs '\n' as tag_separator.
# The other finds all games and tags with both name_value_separator and
# tag_separator as ' ', but finds nothing with both as ''.  Both accept '\n'
# as block_separator rather than '\n\n' which puts a blank line at significant
# points. Neither accepts '', no whitespace, between movetext tokens.  Neither
# needs move numbers or black move indicators.
# Hence the choice of default values.
def get_game_pgn_import_format(
    collected_game,
    name_value_separator=" ",
    tag_separator=" ",
    movetext_separator=" ",
    block_separator="\n",
    line_length=79,
):
    """Construct game score in a PGN import format.

    This method cannot generate text which is identical to internal format
    because the movetext tokens have check and checkmate indicator suffixes
    where appropriate.

    """
    if not isinstance(line_length, int):
        return "".join(
            (
                tag_separator.join(
                    collected_game.get_tags(
                        name_value_separator=name_value_separator
                    )
                ),
                block_separator,
                movetext_separator.join(collected_game.get_movetext()),
                block_separator,
            )
        )
    _attt = _add_token_to_text
    text = []
    length = 0
    for t in collected_game.get_tags(
        name_value_separator=name_value_separator
    ):
        length = _add_token_to_text(
            t, text, line_length, tag_separator, length
        )
    text.append(block_separator)
    length = len(block_separator.split("\n")[-1])
    for t in collected_game.get_movetext():
        if t.startswith("{"):
            s = t.split()
            length = _add_token_to_text(
                s.pop(0), text, line_length, movetext_separator, length
            )
            for w in s:
                length = _add_token_to_text(w, text, line_length, " ", length)
        elif t.startswith("$"):
            length = _add_token_to_text(
                t, text, line_length, movetext_separator, length
            )
        elif t.startswith(";"):
            if len(t) + length >= line_length:
                text.append("\n")
            else:
                text.append(movetext_separator)
            text.append(t)
            length = 0
        elif t == "(":
            length = _add_token_to_text(
                t, text, line_length, movetext_separator, length
            )
        elif t == ")":
            length = _add_token_to_text(
                t, text, line_length, movetext_separator, length
            )
        else:
            length = _add_token_to_text(
                t, text, line_length, movetext_separator, length
            )
    text.append(block_separator)
    return "".join(text)


# Derived from _add_token_to_movetext method in pgn_read.core.game module.
def _add_token_to_text(token, text, line_length, token_separator, length):
    if not length:
        text.append(token)
        return len(token)
    elif len(token) + length >= line_length:
        text.append("\n")
        text.append(token)
        return len(token)
    else:
        text.append(token_separator)
        text.append(token)
        return len(token) + length + len(token_separator)



# topleveltext.py
# Copyright 2021 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Provide methods shared by all classes which display games, repertoires,
selection rules, CQL statements, and run engine commands in Toplevels.

"""


class ToplevelText:

    """Mixin providing methods shared by all classes which display selection
    rules, CQL statements, and run engine commands in Toplevels.

    """

    binding_labels = ()

    # Without this, right-click, Shift F10, and Ctrl F10, in CQL* and Query*
    # toplevels cause AttributeError exceptions.  This just causes a null
    # popup menu to be posted: perhaps difficult to notice.  Maybe a reason to
    # show something is needed.
    # The Engine* classes do not have this problem, perhaps because of the
    # 'Run Engine' option or the absence of the 'main display' options.
    # The Game* and Repertoire* classes have reason for a popup here.
    def create_widget_navigation_submenu_for_popup(self, popup):
        pass


class _ToplevelText:

    """Mixin providing methods shared by at least two of the DeleteText,
    EditText, and ShowText, classes.

    """

    def initialize(self):
        oldview = self.oldview
        self.ui_items_in_toplevels.add(oldview)
        self.parent.wm_title(self.get_title_for_object(self.object))
        self.initialize_item_bindings(oldview)
        self.set_item(oldview, self.object)

    def initialize_item_bindings(self, item):
        self.bind_buttons_to_widget(item.score)
        # item.set_score_pointer_to_score_bindings(False)

    def tidy_on_destroy(self):
        """Clear up after Toplevel destruction."""
        self.ui_items_in_toplevels.discard(self.oldview)
        self.ui_base_table.selection.clear()


class ShowText(_ToplevelText):
    def dialog_ok(self):
        """Close the show record toplevel."""
        if self.ui.database is None:
            if self.ok:
                self.ok.destroy()
                self.ok = None
            self.blockchange = True
            return False
        return super().dialog_ok()


class DeleteText(_ToplevelText):
    def dialog_ok(self):
        """Delete record and return delete action response (True for deleted).

        Check that database is open and is same one as deletion action was
        started.

        """
        if self.ui.database is None:
            self.status.configure(
                text="Cannot delete because not connected to a database"
            )
            if self.ok:
                self.ok.destroy()
                self.ok = None
            self.blockchange = True
            return False
        return super().dialog_ok()


class EditText(_ToplevelText):
    def initialize(self):
        if self.oldview:
            self.ui_items_in_toplevels.add(self.oldview)
            self.set_item(self.oldview, self.oldobject)
        newview = self.newview
        self.ui_items_in_toplevels.add(newview)
        self.set_item(newview, self.newobject)
        self.parent.wm_title(self.get_title_for_object(self.newobject))
        self.initialize_item_bindings(newview)

    def dialog_ok(self):
        """Update record and return update action response (True for updated).

        Check that database is open and is same one as update action was
        started.

        """
        if self.ui.database is None:
            self.status.configure(
                text="Cannot update because not connected to a database"
            )
            if self.ok:
                self.ok.destroy()
                self.ok = None
            self.blockchange = True
            return False
        return super().dialog_ok()

    def tidy_on_destroy(self):
        """Clear up after Toplevel destruction."""
        self.ui_items_in_toplevels.discard(self.oldview)
        self.ui_items_in_toplevels.discard(self.newview)
        self.ui_base_table.selection.clear()



# chessgames.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""View a database of chess games created from data in PGN format.

Run "python -m chesstab.chessgames" assuming chesstab is in site-packages and
Python3.3 or later is the system Python.

PGN is "Portable Game Notation", the standard non-proprietary format for files
of chess game scores.

Sqlite3 via the apsw or sqlite packages, Berkeley DB via the db package, or DPT
via the dpt package, can be used as the database engine.

When importing games while running under Wine it will probably be necessary to
use the sibling module "chessgames_winedptchunk".  The only known reason to run
under Wine is to use the DPT database engine on a platform other than Microsoft
Windows.
"""

if __name__ == "__main__":

    from . import APPLICATION_NAME

    try:
        from solentware_misc.gui.startstop import (
            start_application_exception,
            stop_application,
            application_exception,
        )
    except Exception as error:
        import tkinter.messagebox

        try:
            tkinter.messagebox.showerror(
                title="Start Exception",
                message=".\n\nThe reported exception is:\n\n".join(
                    (
                        "Unable to import solentware_misc.gui.startstop module",
                        str(error),
                    )
                ),
            )
        except:
            pass
        raise SystemExit("Unable to import start application utilities")
    try:
        from .gui.chess import Chess
    except Exception as error:
        start_application_exception(
            error, appname=APPLICATION_NAME, action="import"
        )
        raise SystemExit(" import ".join(("Unable to", APPLICATION_NAME)))
    try:
        app = Chess(allowcreate=True)
    except Exception as error:
        start_application_exception(
            error, appname=APPLICATION_NAME, action="initialise"
        )
        raise SystemExit(" initialise ".join(("Unable to", APPLICATION_NAME)))
    try:
        app.root.mainloop()
    except SystemExit:
        stop_application(app, app.root)
        raise
    except Exception as error:
        application_exception(
            error,
            app,
            app.root,
            title=APPLICATION_NAME,
            appname=APPLICATION_NAME,
        )



# selection.py
# Copyright 2019 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Interface to chess database for selection rules index providing methods
which do not need to know the particular engine used.

A superclass will include a base class for particular database engines.
"""


class Selection:

    """Represent subset of games on file that match a selection rule."""

    def __init__(self, dbhome, dbset, dbname, newrow=None):
        super().__init__(dbhome, dbset, dbname, newrow=newrow)

    def get_selection_rule_games(self, fullposition):
        """Find game records matching selection rule."""
        if not fullposition:
            self.set_recordset(self.dbhome.recordlist_nil(self.dbset))
            return
        self.set_recordset(fullposition)
